import ntpath
import os
import json
from os.path import isfile

from PyQt5.QtWidgets import *

from ..diagram.diagram import Diagram
from .workarea import ScrolledWorkArea

SETTINGS_LAST_OPEN_SECTION = 'diagram'
SETTINGS_LAST_OPEN_OPTION = 'last_open'

last_file_name = ""


class DiagramManager:

    FILE_TYPES = "CV Lab diagram (*.cvlab);;JSON (*.json);;All Files (*)"

    def __init__(self, tabs_container, style_manager):
        self.tabs_container = tabs_container
        self.style_manager = style_manager

    def count(self):
        return self.tabs_container.count()

    def get_open_file_name(self):
        global last_file_name
        main_window = self.tabs_container.parent()
        path, _ = QFileDialog.getOpenFileName(main_window, "Open diagram file", last_file_name, self.FILE_TYPES)
        if path:
            last_file_name = path
        return path

    def get_save_file_name(self):
        global last_file_name
        main_window = self.tabs_container.parent()
        path, _ = QFileDialog.getSaveFileName(main_window, "Save diagram as", last_file_name, self.FILE_TYPES)
        if path:
            last_file_name = path
        return path

    def open_diagram_browse(self):
        # TODO: block the UI for the time of loading?
        diagram_path = self.get_open_file_name()
        self.open_diagram_from_path(diagram_path)

    def open_diagram_from_path(self, path):
        # TODO: block the UI for the time of loading?
        if not path or not isfile(path):
            return
        with open(path, 'r') as fp:
            try:
                encoded = fp.read()
                scrolled_wa = ScrolledWorkArea(Diagram(), self.style_manager)
                base_path = os.path.abspath(path + "/../").replace("\\","/")
                scrolled_wa.load_diagram_from_json(encoded, base_path)
                full_path = os.path.abspath(str(path))
                self.open_diagram(scrolled_wa, full_path)
                scrolled_wa.workarea.actualize_style()
            except Exception as e:
                print("Error: could not load diagram {} - {}".format(path, e))

    def open_from_settings(self, config):
        try:
            files = config.get(SETTINGS_LAST_OPEN_SECTION, SETTINGS_LAST_OPEN_OPTION)
            if files is None:
                return
            paths = json.loads(files)
            for path in paths:
                self.open_diagram_from_path(path)
        except ValueError:
            print("Error loading settings file: wrong entry in the field: [diagram] last_open.")

    def open_diagram(self, scrolled_wa=None, full_path=None):
        # TODO: block the UI for the time of loading?
        to_open = scrolled_wa
        name = None
        if scrolled_wa is None:
            to_open = ScrolledWorkArea(Diagram(), self.style_manager)
        if full_path is None:
            name = "new*"
        else:
            name = get_file_name_from_path(full_path)
            if name.endswith(".cvlab"): name = name[:-6]
        idx = self.tabs_container.addTab(to_open, name)
        if full_path is not None:
            self.tabs_container.setTabToolTip(idx, full_path)
        self.tabs_container.setCurrentIndex(idx)
        for e in to_open.diagram.elements:  # need to do this here, as we want to process repaint events
            e.state_changed.emit()

    def save_to_settings(self, settings):
        count = self.tabs_container.count()
        paths = [str(self.tabs_container.tabToolTip(i)) for i in range(count)]
        entry = json.dumps(paths)
        settings.set(SETTINGS_LAST_OPEN_SECTION, SETTINGS_LAST_OPEN_OPTION, entry)

    def save_diagram_as(self, tab_idx=None):
        path = self.get_save_file_name()
        if path:
            tab_idx = tab_idx if tab_idx is not None else self.tabs_container.currentIndex()
            scrolled_workarea = self.tabs_container.widget(tab_idx)
            diagram = scrolled_workarea.workarea.diagram
            # TODO: block the UI for the time of saving
            save_diagram_to_file(diagram, path)
            diagram_name = get_file_name_from_path(str(path))
            self.tabs_container.setTabText(tab_idx, diagram_name)
            self.tabs_container.setTabToolTip(tab_idx, str(path))

    def save_diagram(self, tab_idx=None):
        tab_idx = tab_idx or self.tabs_container.currentIndex()
        path = self.tabs_container.tabToolTip(tab_idx)
        if not path or path == "":
            self.save_diagram_as(tab_idx)
            return
        diagram = self.tabs_container.widget(tab_idx).diagram
        if diagram:
            save_diagram_to_file(diagram, path)

    def close_diagram(self, tab_idx=None):
        # TODO: block UI?
        if tab_idx is None:
            tab_idx = self.tabs_container.currentIndex()
        scrolled_workarea = self.tabs_container.widget(tab_idx)
        self.tabs_container.removeTab(tab_idx)
        if scrolled_workarea:
            scrolled_workarea.diagram.clear()
            scrolled_workarea.deleteLater()

    def close_all_diagrams(self):
        while self.tabs_container.count():
            self.close_diagram()

    def current_workarea(self):
        tab_idx = self.tabs_container.currentIndex()
        workarea = self.tabs_container.widget(tab_idx)
        return workarea


def get_file_name_from_path(path):
    return os.path.basename(str(path))


def save_diagram_to_file(diagram, path):
    base_path = os.path.abspath(path + "/../").replace("\\","/")
    encoded = diagram.save_to_json(base_path)
    with open(path, 'w') as fp:
        fp.write(encoded)



 # ############################################################################
 #
 # Copyright (c) Microsoft Corporation. 
 #
 # Available under the Microsoft PyKinect 1.0 Alpha license.  See LICENSE.txt
 # for more information.
 #
 # ###########################################################################/




import os
import ctypes
from os import path

_audiodll_path = os.path.join(os.environ['WINDIR'], 'System32', 'KinectAudio10.dll')

_MAX_STR_LEN = 512
_AUDIODLL = ctypes.WinDLL(_audiodll_path)

_audio_path = path.join(path.dirname(__file__), 'PyKinectAudio.dll')
if not os.path.exists(_audio_path):
    _audio_path = path.join(path.dirname(__file__), '..', '..', '..', '..', '..', '..', '..', 'Binaries', 'Debug11.0', 'PyKinectAudio.dll')
    if not path.exists(_audio_path):
        _audio_path = path.join(path.dirname(__file__), '..', '..', '..', '..', '..', '..', '..', 'Binaries', 'Debug10.0', 'PyKinectAudio.dll')
        if not path.exists(_audio_path):
            raise Exception('Cannot find PyKinectAudio.dll')

_PYAUDIODLL = ctypes.CDLL(_audio_path)
_OpenKinectAudio = _PYAUDIODLL.OpenKinectAudio
_OpenKinectAudio.argtypes = [ctypes.c_voidp, ctypes.POINTER(ctypes.c_voidp)]
_OpenKinectAudio.restype = ctypes.HRESULT

_OpenAudioStream = _PYAUDIODLL.OpenAudioStream
_OpenAudioStream.argtypes = [ctypes.c_voidp, ctypes.POINTER(ctypes.c_voidp), ctypes.c_uint32]
_OpenAudioStream.restype = ctypes.HRESULT

_SetDeviceProperty_Bool = _PYAUDIODLL.SetDeviceProperty_Bool
_SetDeviceProperty_Bool.argtypes = [ctypes.c_voidp, ctypes.c_uint32, ctypes.c_uint]
_SetDeviceProperty_Bool.restype = ctypes.HRESULT

_SetDeviceProperty_Int = _PYAUDIODLL.SetDeviceProperty_Int
_SetDeviceProperty_Int.argtypes = [ctypes.c_voidp, ctypes.c_uint32, ctypes.c_int]
_SetDeviceProperty_Int.restype = ctypes.HRESULT

_GetDeviceProperty_Bool = _PYAUDIODLL.GetDeviceProperty_Bool
_GetDeviceProperty_Bool.argtypes = [ctypes.c_voidp, ctypes.c_uint32, ctypes.POINTER(ctypes.c_uint)]
_GetDeviceProperty_Bool.restype = ctypes.HRESULT

_GetDeviceProperty_Int = _PYAUDIODLL.GetDeviceProperty_Int
_GetDeviceProperty_Int.argtypes = [ctypes.c_voidp, ctypes.c_uint32, ctypes.POINTER(ctypes.c_int)]
_GetDeviceProperty_Int.restype = ctypes.HRESULT


#class AecQualityMetrics(ctypes.Structure):
#    _fields_ = [
#                ('timestamp', ctypes.c_longlong),
#                ('convergence_flag', ctypes.c_byte),
#                ('mic_clipped_flag', ctypes.c_byte),
#                ('mic_silence_flag', ctypes.c_byte),
#                ('pstv_feadback_flag', ctypes.c_byte),
#                ('spk_clipped_flag', ctypes.c_byte),
#                ('spk_mute_flag', ctypes.c_byte),
#                ('glitch_flag', ctypes.c_byte),
#                ('double_talk_flag', ctypes.c_byte),
#                ('glitch_count', ctypes.c_ulong),
#                ('mic_clip_count', ctypes.c_ulong),
#                ('duration', ctypes.c_float),
#                ('ts_variance', ctypes.c_float),
#                ('ts_drift_rate', ctypes.c_float),
#                ('voice_level', ctypes.c_float),
#                ('noise_Level', ctypes.c_float),
#                ('echo_return_loss_enhancement', ctypes.c_float),
#                ('avg_echo_return_loss_enhancement', ctypes.c_float),
#                ('reserved', ctypes.c_uint32),
#               ]

#_GetDeviceProperty_QualityMetrics = _PYAUDIODLL.GetDeviceProperty_QualityMetrics
#_GetDeviceProperty_QualityMetrics.argtypes = [ctypes.c_voidp, ctypes.c_uint32, ctypes.POINTER(AecQualityMetrics)]
#_GetDeviceProperty_QualityMetrics.restype = ctypes.HRESULT

_ReadAudioStream = _PYAUDIODLL.ReadAudioStream
_ReadAudioStream.argtypes = [ctypes.c_voidp, ctypes.c_voidp, ctypes.c_uint32, ctypes.POINTER(ctypes.c_uint32)]
_ReadAudioStream.restype = ctypes.HRESULT

_IUnknownRelease = _PYAUDIODLL.IUnknownRelease
_IUnknownRelease.argtypes = [ctypes.c_voidp]
_IUnknownRelease.restype = None

_Recognize_Callback = ctypes.WINFUNCTYPE(None, ctypes.c_wchar_p)

_ReadCallback = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_uint32, ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint32))

_OLE32 = ctypes.WinDLL('ole32.dll')
_CoInitialize = _OLE32.CoInitialize
_CoInitialize.argtypes = [ctypes.c_voidp]
_CoInitialize.restype = ctypes.HRESULT

_CoInitialize(None)

class AudioDeviceInfo(ctypes.Structure):
    """Describes a systems Kinect sensors. """
    _fields_ = [('device_name', ctypes.c_wchar * _MAX_STR_LEN),
                ('device_id', ctypes.c_wchar * _MAX_STR_LEN),
                ('device_index', ctypes.c_int),
                ]

_NuiGetMicrophoneArrayDevices = _AUDIODLL.NuiGetMicrophoneArrayDevices
_NuiGetMicrophoneArrayDevices.argtypes = [ctypes.POINTER(AudioDeviceInfo), ctypes.POINTER(ctypes.c_int)]
_NuiGetMicrophoneArrayDevices.restype = ctypes.HRESULT

def GetKinectDeviceInfo():
    '''returns a sequence of AudioDeviceInfo objects describing the available Kinect audio devices'''
    count = ctypes.c_int()
    _NuiGetMicrophoneArrayDevices(None, 0, ctypes.byref(count))
    deviceArray = (AudioDeviceInfo * count.value)()

    _NuiGetMicrophoneArrayDevices(deviceArray, count.value, ctypes.byref(count))

    return [arrayObj for arrayObj in deviceArray]


_MFPKEY_WMAAECMA_SYSTEM_MODE = 2
_MFPKEY_WMAAECMA_DMO_SOURCE_MODE = 3
_MFPKEY_WMAAECMA_DEVICE_INDEXES = 4
_MFPKEY_WMAAECMA_FEATURE_MODE = 5
_MFPKEY_WMAAECMA_FEATR_FRAME_SIZE = 6
_MFPKEY_WMAAECMA_FEATR_ECHO_LENGTH = 7
_MFPKEY_WMAAECMA_FEATR_NS = 8
_MFPKEY_WMAAECMA_FEATR_AGC = 9
_MFPKEY_WMAAECMA_FEATR_AES = 10
_MFPKEY_WMAAECMA_FEATR_VAD = 11
_MFPKEY_WMAAECMA_FEATR_CENTER_CLIP = 12
_MFPKEY_WMAAECMA_FEATR_NOISE_FILL = 13
_MFPKEY_WMAAECMA_RETRIEVE_TS_STATS = 14
_MFPKEY_WMAAECMA_QUALITY_METRICS = 15
_MFPKEY_WMAAECMA_DEVICEPAIR_GUID = 0x11
_MFPKEY_WMAAECMA_FEATR_MICARR_MODE = 0x12
_MFPKEY_WMAAECMA_FEATR_MICARR_BEAM = 0x13
_MFPKEY_WMAAECMA_MIC_GAIN_BOUNDER = 0x15
_MFPKEY_WMAAECMA_FEATR_MICARR_PREPROC = 20


_AudioPropertySetters = {
        bool : _SetDeviceProperty_Bool,
        int : _SetDeviceProperty_Int,
}

_AudioPropertyGetters = {
        bool : (_GetDeviceProperty_Bool, ctypes.c_uint),
        int : (_GetDeviceProperty_Int, ctypes.c_int),
        #AecQualityMetrics: (_GetDeviceProperty_QualityMetrics,  AecQualityMetrics),
 }

class _AudioSourceProperty(object):
    """internal descriptor used for all of our properties"""
    def __init__(self, index, prop_type, doc = None):
        self.index = index
        self.prop_type = prop_type
        self.__doc__ = doc

    def __get__(self, inst, context):
        getter_func, getter_type = _AudioPropertyGetters[self.prop_type]
        value = getter_type()
        getter_func(inst._dmo, self.index, ctypes.byref(value))
        return value.value


class _SettableAudioSourceProperty(_AudioSourceProperty):
    def __set__(self, inst, value):
        _AudioPropertySetters[self.prop_type](inst._dmo, self.index, value)


class MicArrayMode(object):
    MicArrayAdaptiveBeam = 0x1100,
    MicArrayExternBeam = 0x800,
    MicArrayFixedBeam = 0x400,
    MicArraySimpleSum = 0x100,
    MicArraySingleBeam = 0x200,
    MicArraySingleChan = 0

 
class _AudioFile(object):
    """provides a file-like object for reading from the kinect audio stream"""

    def __init__(self, stream):
        self.closed = False
        self.__ISpStreamFormat__ = stream
        self._buffer = (ctypes.c_byte * 4096)()
        self._buffered_bytes = 0
        self._buffer_start = 0

    def __del__(self):
        if self.__ISpStreamFormat__ is not None:
            _IUnknownRelease(self.__ISpStreamFormat__)

    def close(self):
        self.closed = True
        _IUnknownRelease(self.__ISpStreamFormat__)
        self.__ISpStreamFormat__ = None

    def flush(self):
        pass

    def __next__(self):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()

    def read(self, size = 4096):
        if self.closed:
            raise IOError('Kinect audio source has been closed')

        to_read = size
        res = []
        bytes_read = ctypes.c_uint32()
        while to_read != 0:            
            if self._buffer_start != self._buffered_bytes:
                to_append = min(to_read, self._buffered_bytes - self._buffer_start)

                res.append(ctypes.string_at(ctypes.addressof(self._buffer) + self._buffer_start, to_append))
                
                self._buffer_start += to_append
                to_read -= to_append

                if not to_read:
                    break

            # read more data...
            _ReadAudioStream(self.__ISpStreamFormat__, self._buffer, len(self._buffer), ctypes.byref(bytes_read))
            self._buffer_start = 0
            self._buffered_bytes = bytes_read.value
        
        return ''.join(res)

    def readline(self, size = None):
        raise NotImplementedError()

    def readlines(self, sizehint = None):
        raise NotImplementedError()

    def xreadlines(self):
        raise NotImplementedError()

    def seek(self, offset, whence):
        raise NotImplementedError()

    def tell(self):
        raise NotImplementedError()

    def truncate(self, size = None):
        raise IOError('KinectAudio file not open for writing')

    def write(self, str):
        raise IOError('KinectAudio file not open for writing')

    def writelines(self, sequence):
        raise IOError('KinectAudio file not open for writing')
    
    @property
    def name(self):
        return 'KinectAudio'

    @property
    def mode(self):
        return 'rb'

class KinectAudioSource(object):
    def __init__(self, device = None):
        self._dmo = None
        dmo = ctypes.c_voidp()
        if device is None:
            from pykinect import nui
            
            device = nui.Runtime(nui.RuntimeOptions.uses_audio)
                
        _OpenKinectAudio(device._nui, ctypes.byref(dmo))
        self._dmo = dmo
        self._file = None
        self._device = device

    def __del__(self):
        if self._dmo is not None:
            _IUnknownRelease(self._dmo)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        _IUnknownRelease(self._dmo)
        self._dmo = None
        self._device = None

    def start(self, readStaleThreshold = 500):
        """Starts capturing audio from the Kinect sensor's microphone array into a buffer.   Returns a file-like object that represents the audio stream, which is in 16khz, 16 bit PCM format. 
        
readStaleThreshold: Specifies how long to retain data in the buffer (in milliseconds). If you do not read from the stream for longer than this "stale data" threshold value the DMO discards any buffered audio.

        """
        if self._file is not None:
            raise Exception('Capture already started')

        audio_stream = ctypes.c_voidp()
        _OpenAudioStream(self._dmo, ctypes.byref(audio_stream), readStaleThreshold)

        self._file = _AudioFile(audio_stream)
        return self._file

    def stop(self):
        if self._file is not None:
            self._file.close()        
    
    acoustic_echo_suppression = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATR_AES, int)
    automatic_gain_control = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATR_AGC, bool)
    center_clip = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATR_CENTER_CLIP, bool)
    echo_length = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATR_ECHO_LENGTH, int)
    feature_mode = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATURE_MODE, bool)
    frame_size = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATR_FRAME_SIZE, int)
    gain_bounder = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_MIC_GAIN_BOUNDER, bool)
    mic_array_mode = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATR_MICARR_MODE, int)
    mic_array_preprocess = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATR_MICARR_PREPROC, bool)
    noise_fill = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATR_NOISE_FILL, bool)
    noise_suppression = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATR_NS, int)
    #quality_metrics = _AudioSourceProperty(_MFPKEY_WMAAECMA_QUALITY_METRICS, AecQualityMetrics) 
    retrieve_ts_stats = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_RETRIEVE_TS_STATS, bool)
    source_mode = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_DMO_SOURCE_MODE, bool)
    system_mode = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_SYSTEM_MODE, int)
    voice_activity_detector = _SettableAudioSourceProperty(_MFPKEY_WMAAECMA_FEATR_VAD, int)
    device_pair_guid = _AudioSourceProperty(_MFPKEY_WMAAECMA_DEVICEPAIR_GUID, object) # TODO: Read only, guid type

    # TODO: Properties
    # speaker_index
    # sound_source_position_confidence
    # sound_source_position
    # mic_array_beam_angle

    # TODO: Events
    # beam_changed event

    pass




from distutils.util import strtobool

from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from . import config


class MenuBar(QMenuBar):
    def __init__(self, parent):
        super(MenuBar, self).__init__(parent)
        main_window = parent
        main_window.setMenuBar(self)

        file_menu = self.addMenu('&File')
        file_menu.addAction(NewAction(file_menu, main_window))
        file_menu.addAction(OpenAction(file_menu, main_window))
        file_menu.addAction(SaveAction(file_menu, main_window))
        file_menu.addAction(SaveAsAction(file_menu, main_window))
        file_menu.addAction(CloseDiagramAction(file_menu, main_window))
        file_menu.addSeparator()
        file_menu.addAction(CloseAppAction(file_menu, main_window))

        view_menu = self.addMenu('&View')
        view_menu.addMenu(ColorThemeMenu(view_menu, main_window))
        view_menu.addAction(HighQualityAction(view_menu, main_window))
        view_menu.addAction(LivePreviewsAction(view_menu, main_window))
        view_menu.addAction(PreviewOnTopAction(view_menu, main_window))
        view_menu.addAction(ResetZoomAction(view_menu, main_window))
        view_menu.addAction(ExperimentalElementsAction(view_menu, main_window))

        help_menu = self.addMenu("&Help")
        help_menu.addAction(AboutAction(help_menu, main_window))


class Action(QAction):
    def __init__(self, name, parent, main_window):
        super(Action, self).__init__(name, parent)
        self.main_window = main_window
        self.settings = config.ConfigWrapper.get_settings()


class NewAction(Action):
    def __init__(self, parent, main_window):
        super(NewAction, self).__init__('&New', parent, main_window)
        self.setShortcut(QKeySequence(Qt.CTRL + Qt.Key_N))
        self.triggered.connect(self.open_new)

    @pyqtSlot()
    def open_new(self):
        self.main_window.diagram_manager.open_diagram()


class OpenAction(Action):
    def __init__(self, parent, main_window):
        super(OpenAction, self).__init__('&Open...', parent, main_window)
        self.setShortcut(QKeySequence(Qt.CTRL + Qt.Key_O))
        self.triggered.connect(self.open)

    @pyqtSlot()
    def open(self):
        self.main_window.diagram_manager.open_diagram_browse()


class SaveAsAction(Action):
    def __init__(self, parent, main_window):
        super(SaveAsAction, self).__init__('&Save as...', parent, main_window)
        self.setShortcut(QKeySequence(Qt.CTRL + Qt.SHIFT + Qt.Key_S))
        self.triggered.connect(self.save_as)

    @pyqtSlot()
    def save_as(self):
        self.main_window.diagram_manager.save_diagram_as()


class SaveAction(Action):
    def __init__(self, parent, main_window):
        super(SaveAction, self).__init__('Save', parent, main_window)
        self.setShortcut(QKeySequence(Qt.CTRL + Qt.Key_S))
        self.triggered.connect(self.save)

    @pyqtSlot()
    def save(self):
        self.main_window.diagram_manager.save_diagram()


class CloseDiagramAction(Action):
    def __init__(self, parent, main_window):
        super(CloseDiagramAction, self).__init__('Close diagram', parent, main_window)
        self.setShortcut(QKeySequence(Qt.CTRL + Qt.Key_W))
        self.triggered.connect(self.close)

    @pyqtSlot()
    def close(self):
        self.main_window.diagram_manager.close_diagram()


class CloseAppAction(Action):
    def __init__(self, parent, main_window):
        super(CloseAppAction, self).__init__('E&xit', parent, main_window)
        self.setShortcut(QKeySequence(Qt.ALT + Qt.Key_F4))
        self.triggered.connect(self.exit)

    @pyqtSlot()
    def exit(self):
        self.main_window.diagram_manager.close_all_diagrams()
        self.main_window.application.quit()


class ColorThemeMenu(QMenu):
    def __init__(self, parent, main_window):
        super(ColorThemeMenu, self).__init__("&Set color theme", parent)
        self.main_window = main_window
        self.fill_styles()

    def fill_styles(self):
        styles = self.main_window.style_manager.get_available_stylesheets()
        current = self.main_window.settings.get_with_default(config.VIEW_SECTION, config.STYLE)
        for style in styles:
            action = self.addAction(style)
            action.setCheckable(True)
            if style == current:
                action.setChecked(True)
            else:
                action.setChecked(False)
            action.triggered.connect(self.style_chosen)

    @pyqtSlot()
    def style_chosen(self):
        chosen_style = self.sender()
        self.apply_style(chosen_style.text())

    def apply_style(self, style):
        for action in self.actions():
            if action.text() == style:
                action.setChecked(True)
            else:
                action.setChecked(False)
        self.main_window.style_manager.apply_stylesheet(style)


class HighQualityAction(Action):
    def __init__(self, parent, main_window):
        super(HighQualityAction, self).__init__('&High quality previews', parent, main_window)
        self.setCheckable(True)
        self.value = bool(strtobool(self.settings.get_with_default(config.VIEW_SECTION, config.VIEW_HQ_OPTION)))
        self.setChecked(self.value)
        self.triggered.connect(self.toggle_quality)

    @pyqtSlot()
    def toggle_quality(self):
        self.value = not self.value
        self.setChecked(self.value)
        self.settings.set(config.VIEW_SECTION, config.VIEW_HQ_OPTION, self.value)


class LivePreviewsAction(Action):
    def __init__(self, parent, main_window):
        super(LivePreviewsAction, self).__init__('&Show live results', parent, main_window)
        self.setCheckable(True)
        self.value = bool(strtobool(self.settings.get_with_default(config.VIEW_SECTION,
                                                                   config.LIVE_IMAGE_PREVIEW_OPTION)))
        self.setChecked(self.value)
        self.triggered.connect(self.switch)

    @pyqtSlot()
    def switch(self):
        self.value = not self.value
        self.setChecked(self.value)
        self.settings.set(config.VIEW_SECTION, config.LIVE_IMAGE_PREVIEW_OPTION, self.value)


class PreviewOnTopAction(Action):
    def __init__(self, parent, main_window):
        super(PreviewOnTopAction, self).__init__('&Preview always on top', parent, main_window)
        self.setCheckable(True)
        self.value = bool(strtobool(self.settings.get_with_default(config.VIEW_SECTION, config.PREVIEW_ON_TOP_OPTION)))
        self.setChecked(self.value)
        self.triggered.connect(self.switch)

    @pyqtSlot()
    def switch(self):
        self.value = not self.value
        self.setChecked(self.value)
        self.settings.set(config.VIEW_SECTION, config.PREVIEW_ON_TOP_OPTION, self.value)


class ResetZoomAction(Action):
    def __init__(self, parent, main_window):
        super(ResetZoomAction, self).__init__('Reset zoom', parent, main_window)
        self.setShortcut(QKeySequence(Qt.CTRL + Qt.Key_0))
        self.triggered.connect(self.reset_zoom)

    @pyqtSlot()
    def reset_zoom(self):
        workarea = self.main_window.diagram_manager.current_workarea()
        if workarea:
            workarea.workarea.zoom(1.0)


class ExperimentalElementsAction(Action):
    def __init__(self, parent, main_window):
        super(ExperimentalElementsAction, self).__init__('&Experimental elements', parent, main_window)
        self.setCheckable(True)
        self.value = bool(strtobool(self.settings.get_with_default(config.ELEMENTS_SECTION, config.EXPERIMENTAL_ELEMENTS)))
        self.setChecked(self.value)
        self.triggered.connect(self.toggle)

    @pyqtSlot()
    def toggle(self):
        self.value = not self.value
        self.setChecked(self.value)
        self.settings.set(config.ELEMENTS_SECTION, config.EXPERIMENTAL_ELEMENTS, self.value)
        QMessageBox.information(self.main_window, "Information", "You must restart CV Lab to enable/disable experimental elements.")


class AboutAction(Action):
    message = """\
<h1>CV Lab - Computer Vision Laboratory</h1>
<h3>A rapid prototyping tool for computer vision algorithms</h3>

<p>
Homepage on GitHub: <a href="https://github.com/cvlab-ai/cvlab">https://github.com/cvlab-ai/cvlab</a><br/>  
PyPI package: <a href="https://pypi.python.org/pypi/cvlab">https://pypi.python.org/pypi/cvlab</a>
</p>

<p>
<b>Authors:</b> Adam Brzeski, Jan Cychnerski<br/>
<b>License:</b> AGPL-3.0+<br/><br/>
<i>Copyright 2013-2019</i>
</p>"""

    def __init__(self, parent, main_window):
        super().__init__("About", parent, main_window)
        self.triggered.connect(self.execute)

    @pyqtSlot()
    def execute(self):
        QMessageBox.about(self.main_window, self.main_window.windowTitle(), self.message)



from ..core.core_element import CoreElement
from .connectors import Input, Output

data_header = """\

from collections import defaultdict
from copy import copy
from threading import Lock, RLock


class Data:
    NONE = 0
    SEQUENCE = 1
    IMAGE = 2

    def __init__(self, value=None, _type=IMAGE):
        self._type = _type
        if _type == Data.SEQUENCE and value is None:
            self._value = []
        else:
            self._value = value
        self.lock = RLock()

    def clear(self):
        with self.lock:
            if self._type == Data.SEQUENCE:
                for d in self._value:
                    d.clear()
            else:
                self.value = None

    def copy(self):
        with self.lock:
            if self._type == self.NONE:
                return EmptyData()
            if self._type == self.SEQUENCE:
                return Sequence([d.copy() for d in self._value])
            elif self._type == self.IMAGE:
                if self._value is None or (hasattr(self._value, "size") and not len(self._value)):
                    pass
                    # print("Copying null data!")
                return ImageData(self._value)
            else:
                raise TypeError("Wrong Data.type")

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        with self.lock:
            if self._value is new_value:
                return
            self._value = new_value

    def assign(self, other):
        assert isinstance(other, Data)
        assert self.is_compatible(other)  # todo: assert czy raise?
        if self._type == Data.SEQUENCE:
            assert len(self._value) == len(other._value)  # todo: jak wyĹĽej
            for mine, her in zip(self._value, other._value):
                mine.assign(her)
        else:
            self.value = other.value

    def is_compatible(self, other):
        assert isinstance(other, Data)
        if self._type != other._type: return False
        if self._type == Data.SEQUENCE:
            if len(self._value) != len(other._value): return False
            return all(mine.is_compatible(her) for mine, her in zip(self._value, other._value))
        return True

    def type(self):
        with self.lock:
            if self._value is None or (self._type == Data.IMAGE and hasattr(self.value, "size") and not self._value.size):
                return Data.NONE
            else:
                return self._type

    def sequence_get_value(self, sequence_number):
        with self.lock:
            if self.type() != Data.SEQUENCE:
                return self
            else:
                return self._value[min(sequence_number, len(self._value) - 1)]

    __getitem__ = sequence_get_value

    def __iter__(self):
        with self.lock:
            if self.type() == Data.NONE:
                return iter([])
            elif self.type() == Data.SEQUENCE:
                return iter(self._value)
            else:
                return iter([self])

    def desequence_all(self):
        with self.lock:
            if self._type == Data.NONE: return [None]
            if self._type == Data.IMAGE: return [self._value]
            if self._type == Data.SEQUENCE:
                t = []
                for d in self._value:
                    t += d.desequence_all()
                return t
            raise TypeError("Wrong data type - cannot desequence")

    def is_complete(self):
        with self.lock:
            t = self.type()
            if t == Data.NONE:
                return False
            elif t == Data.IMAGE:
                return True
            else:
                return all(d.is_complete() for d in self._value)

    def create_placeholder(self):
        if self._type == Data.SEQUENCE:
            return Data([d.create_placeholder() for d in self._value], Data.SEQUENCE)
        else:
            return Data()

    def __str__(self):
        return repr(self)

    def __repr__(self):
        id_ = "0x{:08X}".format(id(self))
        if self._type == Data.NONE: return '<Data [empty] at {}>'.format(id_)
        if self.type() == Data.IMAGE:
            try:
                shape = "?"
                shape = len(self._value)
                shape = self._value.shape
            except Exception:
                pass
            return '<Data [image {}] at {}>'.format(shape, id_)
        if self.type() == Data.NONE: return "<Data [empty image] at {}>".format(id_)
        if self.type() == Data.SEQUENCE:
            s = ""
            images_count = 0
            none_count = 0
            for d in self._value:
                if d.type() == Data.SEQUENCE:
                    s += str(d) + ", "
                if d.type() == Data.IMAGE:
                    images_count += 1
                if d.type() == Data.NONE:
                    none_count += 1
            if images_count:
                s += str(images_count) + " images, "
            if none_count:
                s += str(none_count) + " nones, "
            if len(s) > 0:
                s = s.rstrip(', ')
            s = '<Sequence ' + '[' + s + '] at {}>'.format(id_)
            return s
        raise TypeError("Wrong data type - cannot desequence")

    def __nonzero__(self):
        return self.type() != Data.NONE

    ready = __nonzero__

    def __eq__(self, other):
        if not isinstance(other, Data): return False
        with self.lock, other.lock:
            if self.type() != other.type(): return False
            if self.type() == Data.SEQUENCE:
                if len(self._value) != len(other._value): return False
                assert all(isinstance(d, Data) for d in self._value)
                assert all(isinstance(d, Data) for d in other._value)
                return all(a == b for a, b in zip(self._value, other._value))
            else:
                return self._value is other._value
                # todo: if we want logical equality rather than reference equality, we shall use this:
                # elif self.type() == Data.IMAGE:
                # #assert isinstance(self._value, np.ndarray)
                # return self._value is other._value and np.array_equal(self._value, other._value)
                # else:
                #     raise ValueError("Wrong data type")


class EmptyOptionalData(Data):
    def ready(self):
        return True


def Sequence(values=None):
    return Data(values, Data.SEQUENCE)


def EmptyData():
    return Data(None, Data.NONE)


def ImageData(value=None):
    return Data(value, Data.IMAGE)

"""


file_template = """\
#!/usr/bin/env python
#
# Code generated with CV Lab
# https://github.com/cvlab-ai/cvlab
#


### imports ###

{imports}

### helpers ###

{helpers}

### functions ###

{functions}

### process function ###

{process_function}

### script ###

if __name__ == "__main__":
    import sys
    if len(sys.argv) <= 1:
        print("Usage: generated.py {inputs}")
        exit(1)
    args = []
    for arg in sys.argv[1:]:
        args.append(Data(cv.imread(arg)))
    outputs = process(*args)
    for name, output in outputs.items():
        cv.imwrite("output-{{}}.png".format(name), output.value)
        
"""


class CodeGenerator:
    def __init__(self, element):
        self.base_element = element
        self.functions = {}  # code -> name
        self.elem_names = {}  # element -> name
        self.imports = {"import numpy as np", "import cv2 as cv", "import cv2"}
        self.inputs = set()
        self.output = ""
        self.elem_cnt = 0
        self.code = ""

    def generate(self):
        self.process(self.base_element)

        imports = "\n".join(self.imports)

        helpers = data_header

        functions = "\n\n".join(self.functions)

        inputs = ", ".join(self.inputs)
        code = self.code.replace("\n", "\n    ")
        output = self.output
        process_function = "def process({inputs}):\n    {code}\n    return {output}\n".format(**locals())

        return file_template.format(**locals())

    def process(self, element):
        assert isinstance(element, CoreElement)

        # fixme: this may not work for elements with inputs set as multiple

        self.elem_cnt += 1
        element_name = element.__class__.__name__.lower() + str(self.elem_cnt)
        func_name, func_code, input_names = element.get_source()
        self.inputs.update(["{element_name}_{input_name}".format(element_name=element_name, input_name=input_name) for input_name in input_names])
        if func_code:
            self.functions[func_code] = func_name
        if element is self.base_element:
            self.output = element_name
        inputs = {}  # input name -> value
        for input in element.inputs.values():
            assert isinstance(input, Input)
            for output in input.connected_from:
                assert isinstance(output, Output)
                in_elem = output.parent
                if in_elem not in self.elem_names:
                    self.elem_names[in_elem] = self.process(in_elem)
                value = '{}["{}"]'.format(self.elem_names[in_elem], output.name)
                if input.multiple:
                    multi_number = len(inputs)
                    inputs[input.name + str(multi_number)] = value
                else:
                    inputs[input.name] = value
        if func_code:
            inputs = "{" + ",".join(['"{}":{}'.format(input_name, value) for input_name, value in inputs.items()]) + "}"
            params = "{" + ",".join(['"{}":{}'.format(param_name, repr(value.get())) for param_name, value in element.parameters.items()]) + "}"
            self.code += """\
{element_name} = {{}}
{func_name}({inputs}, {element_name}, {params})
""".format(**locals())
        else:
            inputs = "{" + ",".join(['"{output_name}":{element_name}_{output_name}'.format(output_name=output_name, element_name=element_name) for output_name in input_names]) + "}"
            self.code += "{element_name} = {inputs}\n".format(**locals())

        return element_name


def generate(element):
    return CodeGenerator(element).generate()




from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from cvlab.view.spin_widget import *
from ..diagram.parameters import *
from .highlighter import Highlighter


class GuiBaseParameter(QHBoxLayout):
    def __init__(self, parameter):
        super().__init__()
        self.parameter = parameter
        self.setContentsMargins(0,0,0,0)
        self.setSpacing(2)


class GuiButtonParameter(GuiBaseParameter):
    def __init__(self, parameter):
        super().__init__(parameter)
        self.button = QPushButton(self.parameter.name)
        self.button.setObjectName("ButtonParameterButton")
        self.button.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.button.clicked.connect(self.clicked)
        self.addWidget(self.button)

    @pyqtSlot()
    def clicked(self):
        self.parameter.clicked()


class GuiPathParameter(GuiBaseParameter):
    def __init__(self, parameter):
        super().__init__(parameter)
        self.label = QLabel(self.parameter.name)
        self.label.setObjectName("PathParameterName")
        self.path = QLineEdit()
        self.path.setObjectName("PathParameterValue")
        self.path.setEnabled(False)
        self.path.setToolTip(self.path.text())
        self.browse = QPushButton("...")
        self.browse.setObjectName("PathParameterButton")
        self.browse.clicked.connect(self.choose_path)
        self.addWidget(self.label)
        self.addWidget(self.path)
        self.addWidget(self.browse)
        self.parameter.value_changed.connect(self.on_value_changed)
        self.on_value_changed()

    @pyqtSlot()
    def on_value_changed(self):
        self.set_path_(self.parameter.get())

    def set_path_(self, path):
        if path != "":
            self.path.setText(path)
            self.path.setToolTip(path)
            self.parameter.set(str(path))

    @pyqtSlot()
    def choose_path(self):
        if self.parameter.save_mode:
            path, _ = QFileDialog.getSaveFileName(self.browse, "Save file...", self.parameter.get())
        else:
            path, _ = QFileDialog.getOpenFileName(self.browse, "Open file...", self.parameter.get())
        self.set_path_(str(path))


class GuiMultiPathParameter(GuiPathParameter):
    def __init__(self, parameter):
        super().__init__(parameter)

    @pyqtSlot()
    def on_value_changed(self):
        self.set_paths_(self.parameter.get())

    def set_paths_(self, paths):
        if len(paths) > 0:
            # TODO: put some images list instead of ...
            self.path.setText("...")
            self.path.setToolTip("...")
            self.parameter.set(paths)

    @pyqtSlot()
    def choose_path(self):
        paths, _ = QFileDialog.getOpenFileNames(self.browse, "Open multiple files...", self.parameter.get()[0])
        self.set_paths_([str(p) for p in paths])


class GuiDirectoryParameter(GuiPathParameter):
    @pyqtSlot()
    def choose_path(self):
        path = QFileDialog.getExistingDirectory(self.browse, "Open directory...", self.parameter.get())
        self.set_path_(str(path))


class GuiTextParameter(GuiBaseParameter):
    def __init__(self, parameter, element):
        super().__init__(parameter)
        assert isinstance(parameter, TextParameter)
        self.element = element
        self.highlighter = None

        self.label = QLabel(self.parameter.name)
        self.label.setObjectName("TextParameterName")

        self.button = QPushButton("Edit code...")
        self.button.setObjectName("TextParameterButton")
        self.button.clicked.connect(self.edit_code)
        self.addWidget(self.label)
        self.addWidget(self.button)

        self.wnd = QDialog(self.element)
        self.wnd.setModal(False)
        self.wnd.setLayout(QVBoxLayout())
        self.wnd.setObjectName("CodeDialog")
        self.wnd.setWindowTitle(parameter.window_title)
        desktop = QApplication.instance().desktop()
        self.wnd.resize(desktop.screenGeometry(desktop.screenNumber(self.element)).width() // 2,
                        desktop.screenGeometry(desktop.screenNumber(self.element)).height() // 2)
        self.wnd.finished.connect(self.actualize)
        self.wnd_geometry = None

        if parameter.window_content:
            self.wnd.layout().addWidget(QLabel(parameter.window_content))

        self.textedit = QPlainTextEdit()
        self.textedit.setLineWrapMode(self.textedit.NoWrap)
        self.textedit.setWordWrapMode(QTextOption.NoWrap)
        if parameter.live:
            self.textedit.textChanged.connect(self.actualize)
        self.wnd.layout().addWidget(self.textedit)

        ok_button = QPushButton()
        ok_button.setText("OK")
        ok_button.clicked.connect(self.close_but_press)
        self.wnd.layout().addWidget(ok_button)

    @pyqtSlot()
    def edit_code(self):
        self.textedit.setPlainText(self.parameter.get())
        tab_width = QFontMetrics(self.textedit.font()).width("    ")
        self.textedit.setTabStopWidth(tab_width)
        self.highlighter = Highlighter(self.textedit.document())

        if self.wnd_geometry:
            self.wnd.setGeometry(self.wnd_geometry)

        self.wnd.show()

    @pyqtSlot()
    def actualize(self):
        self.parameter.set(str(self.textedit.toPlainText()))

    @pyqtSlot()
    def close_but_press(self):
        self.wnd_geometry = self.wnd.geometry()
        self.wnd.accept()


class GuiIntParameter(GuiBaseParameter):
    def __init__(self, parameter, element):
        super().__init__(parameter)
        self.element = element
        self.label = QLabel(self.parameter.name)
        self.addWidget(self.label)

        self.slider = QSlider(QtCore.Qt.Horizontal)
        self.slider.setTickPosition(QSlider.TicksBelow)
        self.slider.setVisible(False)
        self.slider.setRange(parameter.min, parameter.max)
        self.slider.setSingleStep(parameter.step)
        self.slider.setValue(parameter.get())
        self.slider.valueChanged.connect(self.gui_value_changed)
        self.addWidget(self.slider)
        element.param_sliders.append(self.slider)

        self.spin = SpinBoxEx()
        self.spin.setRange(parameter.min, parameter.max)
        self.spin.setSingleStep(parameter.step)
        self.spin.setValue(parameter.get())
        self.spin.valueChanged.connect(self.gui_value_changed)
        self.addStretch()
        self.addWidget(self.spin)

        self.parameter.value_changed.connect(self.on_value_changed)

    @pyqtSlot(int)
    def gui_value_changed(self, value):
        if self.spin.value() != value or self.slider.value() != value:
            self.parameter.set(int(value))

    @pyqtSlot()
    def on_value_changed(self):
        value = self.parameter.get()
        if self.spin.value() != value:
            self.spin.setValue(value)
        if self.slider.value() != value:
            self.slider.setValue(value)


class GuiFloatParameter(GuiBaseParameter):
    def __init__(self, parameter, element):
        super().__init__(parameter)
        self.element = element
        self.changing = False

        self.label = QLabel(self.parameter.name)
        self.addWidget(self.label)

        self.slider = QSlider(QtCore.Qt.Horizontal)
        self.slider.setTickPosition(QSlider.TicksBelow)
        self.slider.setVisible(False)
        self.slider.setRange(0, 1000)
        self.slider.setSingleStep(1)
        self.slider.setValue(self.value_to_slider(parameter.get()))
        self.slider.valueChanged.connect(self.slider_changed)
        self.addWidget(self.slider)
        element.param_sliders.append(self.slider)

        self.spin = DoubleSpinBoxEx()
        self.spin.setRange(parameter.min, parameter.max)
        self.spin.setSingleStep(parameter.step)
        self.spin.setDecimals(6)
        self.spin.setValue(parameter.get())
        self.spin.valueChanged.connect(self.spin_changed)
        self.spin.stepBy = self.stepBy
        self.addWidget(self.spin)

        self.parameter.value_changed.connect(self.on_value_changed)
        self.ignore_changes = False

    def stepBy(self, steps):
        shift_pressed = (int(QApplication.keyboardModifiers()) & QtCore.Qt.ShiftModifier) != 0
        if shift_pressed:
            QDoubleSpinBox.stepBy(self.spin, steps * 100)
        else:
            QDoubleSpinBox.stepBy(self.spin, steps)

    @pyqtSlot()
    def on_value_changed(self):
        value = self.parameter.get()
        if value != self.slider_to_value(self.slider.value()):
            self.slider.setValue(self.value_to_slider(value))
        if value != self.spin.value():
            self.spin.setValue(value)

    def value_to_slider(self, value):
        v = float(value - self.parameter.min) / (self.parameter.max - self.parameter.min)
        return round(v * 1000)

    def slider_to_value(self, value):
        return value * 0.001 * (self.parameter.max - self.parameter.min) + self.parameter.min

    @pyqtSlot(int)
    def slider_changed(self, value):
        if self.ignore_changes: return
        self.ignore_changes = True
        slider_value = self.slider_to_value(value)
        if slider_value != self.parameter.get():
            self.parameter.set(slider_value)
        self.ignore_changes = False

    @pyqtSlot(float)
    def spin_changed(self, value):
        if self.ignore_changes: return
        self.ignore_changes = True
        if value != self.parameter.get():
            self.parameter.set(value)
        self.ignore_changes = False


class GuiMultiNumberParameter(GuiBaseParameter):
    def __init__(self, parameter, element, count, type):
        super().__init__(parameter)

        self.element = element
        self.count = count
        self.type = type

        self.label = QLabel(self.parameter.name)
        self.addWidget(self.label)

        min_ = self.parameter.min
        max_ = self.parameter.max

        self.spins = []
        for i in range(count):
            spin = SpinBoxEx()
            spin.setRange(min_, max_)
            self.addWidget(spin)
            self.spins.append(spin)

        self.on_value_changed()

        # event is generated instantly, so the whole object must be ready
        for spin in self.spins:
            spin.valueChanged.connect(self.spin_value_changed)

        self.parameter.value_changed.connect(self.on_value_changed)

    def gui_value(self):
        return tuple(self.type(spin.value()) for spin in self.spins)

    @pyqtSlot()
    def spin_value_changed(self):
        parameter_value = self.parameter.get()
        spin_value = self.gui_value()
        if parameter_value != spin_value:
            self.parameter.set(spin_value)

    @pyqtSlot()
    def on_value_changed(self):
        value = self.parameter.get()
        gui_value = self.gui_value()
        if value != gui_value:
            for spin, value in zip(self.spins, value):
                spin.setValue(value)


class GuiComboboxParameter(GuiBaseParameter):
    def __init__(self, parameter, element):
        super().__init__(parameter)
        self.element = element

        self.label = QLabel(self.parameter.name)
        self.addWidget(self.label)

        self.combobox = QComboBox()
        self.combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.combobox.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Ignored)
        self.combobox.setContentsMargins(0,0,0,0)
        self.combobox.setMinimumContentsLength(1)
        self.combobox.view().setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

        for text, value in parameter.values.items():
            self.combobox.addItem(text, value)

        self.combobox.currentIndexChanged.connect(self.combobox_value_changed)
        self.addWidget(self.combobox)

        self.parameter.value_changed.connect(self.on_value_changed)
        self.on_value_changed()

    def gui_value(self):
        value = self.combobox.itemData(self.combobox.currentIndex())
        if hasattr(value, "toPyObject"): value = value.toPyObject()
        return value

    def is_value_outdated(self):
        return self.parameter.get() != self.gui_value()

    @pyqtSlot()
    def combobox_value_changed(self):
        if self.is_value_outdated():
            self.parameter.set(self.gui_value())

    @pyqtSlot()
    def on_value_changed(self):
        if self.is_value_outdated():
            #index = self.combobox.findData(QtCore.QVariant(self.parameter.get()))
            index = self.find_index(self.parameter.get())
            self.combobox.setCurrentIndex(index)

    def find_index(self, data):
        for i in range(self.combobox.count()):
            d = self.combobox.itemData(i)
            if hasattr(d, "toPyObject"): d = d.toPyObject()
            if d == data: return i
        else:
            return -1

#
# class GuiMatrixParameter(GuiBaseParameter):
#
#     class Button(QLabel):
#         pixel_size = (64, 64)
#
#         def __init__(self, parameter, x, y, step):
#             QLabel.__init__(self)
#             self.parameter = parameter
#             self.matrix_x = x
#             self.matrix_y = y
#             self.step = step
#             self.setFixedSize(self.pixel_size[0], self.pixel_size[1])
#             self.setAlignment(QtCore.Qt.AlignCenter | QtCore.Qt.AlignVCenter)
#
#         def mousePressEvent(self, event):
#             if event.button() == QtCore.Qt.LeftButton:
#                 delta = self.step
#             elif event.button() == QtCore.Qt.RightButton:
#                 delta = -self.step
#             else:
#                 event.ignore()
#                 return
#             event.accept()
#             self.add_value(delta)
#
#         def mouseReleaseEvent(self, event):
#             if event.button() in (QtCore.Qt.LeftButton, QtCore.Qt.RightButton):
#                 event.accept()
#             else:
#                 event.ignore()
#
#         def wheelEvent(self, event):
#             event.accept()
#             delta = self.step if event.delta() > 0 else -self.step
#             self.add_value(delta)
#
#         def add_value(self, delta):
#             self.element.matrix[self.matrix_y,self.matrix_x] += delta
#             self.setText(str(self.element.matrix[self.matrix_y,self.matrix_x]))
#             self.element.recalculate(False, False, False, force_units_recalc=True)
#
#     def __init__(self, parameter, element):
#         super(GuiMatrixParameter, self).__init__()
#         assert isinstance(parameter, TextParameter)
#         self.parameter = parameter
#         self.element = element
#
#         self.label = QLabel(self.parameter.name)
#         self.label.setObjectName("MatrixParameterName")
#
#         self.button = QPushButton("Edit matrix...")
#         self.button.setObjectName("MatrixParameterButton")
#         self.button.clicked.connect(self.edit_code)
#         self.addWidget(self.label)
#         self.addWidget(self.button)
#
#         self.wnd = QDialog(self.element)
#         self.wnd.setModal(True)
#         self.wnd.setLayout(QVBoxLayout())
#         self.wnd.setObjectName("MatrixDialog")
#         self.wnd.setWindowTitle(parameter.window_title)
#         self.wnd.finished.connect(self.actualize)
#         self.wnd_geometry = None
#
#         self.size = (7,7)
#
#         self.edit_widget = QWidget()
#         self.edit_widget.setLayout(QGridLayout())
#         self.recreate_editor()
#
#         ok_button = QPushButton()
#         ok_button.setText("OK")
#         ok_button.clicked.connect(self.close_but_press)
#         self.wnd.layout().addWidget(ok_button)
#
#     def recreate_editor(self):
#         QWidget().setLayout(self.edit_widget.layout())
#         layout = QGridLayout()
#         self.edit_widget.setLayout(layout)
#         w, h = self.last_parameters["size"]
#         dtype = self.last_parameters["type"]
#         step = self.last_parameters["step"]
#         self.matrix = np.zeros((h, w), dtype=dtype)
#         layout.setSpacing(0)
#         for x, y in itertools.product(xrange(w), xrange(h)):
#             pixel = self.Button(self, x, y, step)
#             pixel.setText("0")
#             layout.addWidget(pixel, y, x)
#
#     @pyqtSlot()
#     def edit_code(self):
#         self.textedit.setPlainText(self.parameter.get())
#         tab_width = QFontMetrics(self.textedit.font()).width("    ")
#         self.textedit.setTabStopWidth(tab_width)
#         if self.wnd_geometry:
#             self.wnd.setGeometry(self.wnd_geometry)
#         self.wnd.show()
#
#     @pyqtSlot()
#     def actualize(self):
#         self.parameter.set(unicode(self.textedit.toPlainText()))
#
#     @pyqtSlot()
#     def close_but_press(self):
#         self.wnd_geometry = self.wnd.geometry()
#         self.wnd.accept()
#
#



import os
import configparser


SETTINGS_FILENAME = 'settings.cfg'

VIEW_SECTION = 'view'
VIEW_HQ_OPTION = 'high_quality'
LIVE_IMAGE_PREVIEW_OPTION = 'live_preview'
PREVIEW_ON_TOP_OPTION = 'preview_on_top'
STYLE = 'style'

ELEMENTS_SECTION = 'elements'
EXPERIMENTAL_ELEMENTS ='experimental'

UPDATES_SECTION = "updates"
UPDATE_DONT_REMIND_VERSION = "dont_remind_version"

DEFAULTS = {
    VIEW_SECTION: {
        VIEW_HQ_OPTION: 'False',
        LIVE_IMAGE_PREVIEW_OPTION: 'True',
        PREVIEW_ON_TOP_OPTION: 'True',
        STYLE: 'default',
    },
    ELEMENTS_SECTION: {
        EXPERIMENTAL_ELEMENTS: 'False',
    },
    UPDATES_SECTION: {
        UPDATE_DONT_REMIND_VERSION: "0"
    },
}


class ConfigWrapper(configparser.SafeConfigParser):

    cfg = None

    def get_with_default(self, section, name):
        value = self.get(section, name)
        if value is not None:
            return value
        return DEFAULTS[section][name]

    def get(self, section, option, **kwargs):
        try:
            value = configparser.SafeConfigParser.get(self, section, option, **kwargs)
        except (configparser.NoOptionError, configparser.NoSectionError):
            value = None
        return value

    def set(self, section, option, value=None):
        if section not in self.sections():
            self.add_section(section)
        configparser.SafeConfigParser.set(self, section, option, str(value))
        self._save_settings()

    def remove_option(self, section, option):
        configparser.SafeConfigParser.remove_option(self, section, option)
        self._save_settings()

    def _save_settings(self):
        # Create directory if not exists
        settings_path = self.get_settings_path()
        settings_dir = os.path.dirname(settings_path)
        if not os.path.exists(settings_dir):
            os.makedirs(settings_dir)
        # Save settings
        with open(settings_path, 'w') as configfile:
            self.write(configfile)

    @classmethod
    def get_settings(cls):
        if ConfigWrapper.cfg is None:
            ConfigWrapper.cfg = ConfigWrapper()
            ConfigWrapper.cfg.read(cls.get_settings_path())
        return ConfigWrapper.cfg

    @classmethod
    def get_settings_path(cls):
        # For Windows
        if os.name == 'nt':
            return os.path.join(os.environ['appdata'], 'CVLab', SETTINGS_FILENAME)
        # For Unix
        else:
            return os.path.expanduser(os.path.join('~', '.cvlab', SETTINGS_FILENAME))



from collections import deque
from datetime import datetime

from .base import *


class DelayLine(NormalElement):
    name = "Delay line"
    comment = "Delays its input (useful for video processing)"
    num_outputs = 5

    def __init__(self):
        super(DelayLine, self).__init__()
        self.memory = deque(maxlen=5)
        for i in range(self.num_outputs):
            self.memory.append(EmptyData())

    def get_attributes(self):
        return [Input("input")], \
               [Output("o" + str(i + 1)) for i in range(self.num_outputs)], \
            []

    def process_inputs(self, inputs, outputs, parameters):
        for i in range(self.num_outputs):
            outputs["o" + str(i + 1)] = self.memory[self.num_outputs - i - 1]
        self.memory.append(inputs['input'].copy())



class Snapshot(NormalElement):
    name = "Snapshot"
    comment = "Saves actual input (optionally makes a sequence from them)"

    def __init__(self):
        super(Snapshot, self).__init__()
        self.do_snap = False

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [ButtonParameter("snap", self.snap)]

    def get_processing_units(self, inputs, parameters):
        outputs = {"output": inputs["input"].create_placeholder()}
        units = [ProcessingUnit(self, inputs, parameters, outputs)]
        return units, outputs

    def process_units(self):
        if not self.do_snap:
            return
        self.do_snap = False
        unit = self.units[0]
        unit.outputs["output"].assign(unit.inputs["input"])

    def snap(self):
        self.do_snap = True
        self.recalculate(False, False, True)



class Accumulator(NormalElement):
    name = "Accumulator"
    comment = "Accumulates its input at given speed"

    def __init__(self):
        super(Accumulator, self).__init__()
        self.memory = None

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [ComboboxParameter("function",{"Average":"avg","Minimum":"min","Maximum":"max"}),
                FloatParameter("speed", value=0.1, min_=0, max_=1, step=0.01),
                ButtonParameter("reset", self.reset_memory)]

    def process_inputs(self, inputs, outputs, parameters):
        image = inputs["input"].value
        speed = parameters["speed"]
        func = parameters["function"]
        memory = self.memory

        if memory is None or memory.shape != image.shape or memory.dtype != image.dtype:
            memory = image
        else:
            if func == "avg":
                memory = cv.addWeighted(memory, 1-speed, image, speed, 0)
            elif func == "min":
                memory = np.minimum(memory, image)
            elif func == "max":
                memory = np.maximum(memory, image)

        self.may_interrupt()
        self.memory = memory
        outputs["output"] = Data(memory)

    def reset_memory(self):
        self.memory = None




class FpsCounter(NormalElement):
    name = "FPS counter"
    comment = "Counts input frames per second"

    DEQUE_SIZE = 5

    def __init__(self):
        super(FpsCounter, self).__init__()
        self.last_time = datetime.now()
        self.memory = deque(maxlen=5)
        self.bg = False

    def get_attributes(self):
        return [Input("input")], [Output("output")], [IntParameter("memory", value=5, min_=1, max_=50)]

    def get_processing_units(self, inputs, parameters):
        outputs = {"output": Data()}
        return [ProcessingUnit(self, inputs, parameters, outputs)], outputs

    def process(self):
        self.prepare_data()
        if not self.inputs["input"].get(): return
        self.bg = not self.bg
        act_time = datetime.now()
        fps = 1. / (act_time - self.last_time).total_seconds()
        img = np.zeros((30, 50), np.uint8)
        memsize = self.parameters["memory"].get()
        if memsize > 1:
            if self.memory.maxlen != memsize:
                self.memory = deque(self.memory, memsize)
            self.memory.append(fps)
            fps = sum(self.memory) / len(self.memory)
        txt = str(int(round(fps)))
        if self.bg:
            txt += "."
        cv.putText(img, txt, (10, 20), 1, 1, 255, 2)
        if not self.outputs["output"].get():
            self.outputs["output"].put(Data(img))
        else:
            self.outputs["output"].get().value = img
        self.last_time = act_time

register_elements_auto(__name__, locals(), "Video", 4)



import os
from glob import glob

from PyQt5.QtWidgets import QAction

from cvlab.diagram.elements import add_plugin_callback


class OpenExampleAction(QAction):
    def __init__(self, parent, path):
        super().__init__(parent)
        name = os.path.basename(path).replace(".cvlab","").title()
        self.setText(name)
        self.path = path
        self.triggered.connect(self.open)

    def open(self):
        self.parent().diagram_manager.open_diagram_from_path(self.path)


def add_samples(main_window):
    samples = glob(os.path.dirname(__file__) + "/*.cvlab")
    samples.sort()

    print("Adding {} sample diagrams to main menu".format(len(samples)))

    menu = main_window.menuBar()

    samples_menu = menu.addMenu('E&xamples')

    for sample in samples:
        samples_menu.addAction(OpenExampleAction(main_window, sample))


add_plugin_callback(add_samples)




from .base import *


class OpenCVBlur(NormalElement):
    name = "Blur transform"
    comment = "Simple blurring of the image"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [IntParameter("ratio", "Blur ratio", min_=1, max_=255)]

    def process_inputs(self, inputs, outputs, parameters):
        ratio = parameters["ratio"]
        outputs["output"] = Data(cv.blur(inputs["input"].value, (ratio, ratio)))



class GaussianBlur3D(NormalElement):
    name = 'Gaussian blur 3D'
    comment = 'Gaussian blur for 3D data'

    def get_attributes(self):
        return [Input('src', 'src')], \
               [Output('dst', 'dst')], \
               [IntParameter('kernel', None, 9, 1, 999, 2),
                FloatParameter('sigmaX', 'sigmaX', min_=0),
                FloatParameter('sigmaY', 'sigmaY', min_=0),
                FloatParameter('sigmaZ', 'sigmaZ', min_=0),
                IntParameter('borderType', 'borderType')]

    def process_inputs(self, inputs, outputs, parameters):
        image = inputs['src'].value.copy()
        kernel = parameters['kernel']
        sigmaX = parameters['sigmaX']
        sigmaY = parameters['sigmaY']
        sigmaZ = parameters['sigmaZ']
        borderType = parameters['borderType']

        for z in image:
            self.may_interrupt()
            cv.GaussianBlur(z, (kernel, kernel), sigmaX, z, sigmaY, borderType)

        for y in range(image.shape[1]):
            self.may_interrupt()
            xz = image[:, y]
            cv.GaussianBlur(xz, (1, kernel), 0, xz, sigmaZ, borderType)

        outputs['dst'] = Data(image)


register_elements_auto(__name__, locals(), "Filters", 5)




import os
from PyQt5.QtCore import QTimer, pyqtSignal, QObject
from PyQt5.QtWidgets import *

from .. import CVLAB_DIR
from .wires import WireStyle
from . import config


AUTO_REFRESH_STYLESHEET = 0     # Use 1 when working on stylesheets (warning: option causes some artifacts, like flickering
                                # menu bar and disappearing stylesheet when opening File Dialog)


class StyleManager(QObject):

    style_changed = pyqtSignal()

    is_highdpi = QApplication.desktop().screenGeometry().width() > 2500

    def __init__(self, main_window):
        super(StyleManager, self).__init__()
        self.styles_dir = CVLAB_DIR + "/styles/themes"
        self.main_window = main_window
        self.wire_style = None
        self.stylesheet = ""
        self.apply_default_stylesheet()

        # Only for developing stylesheets
        if AUTO_REFRESH_STYLESHEET:
            self.timer = QTimer()
            self.timer.timeout.connect(self.apply_default_stylesheet)
            self.timer.start(500)

    def apply_stylesheet(self, style_name):
        style_name = str(style_name)
        selected_style_path = os.path.join(self.styles_dir, style_name, style_name + ".stylesheet")
        common_style_path = os.path.normpath(self.styles_dir + "/common.stylesheet")
        highdpi_style_path = os.path.normpath(self.styles_dir + "/highdpi.stylesheet")
        stylesheet = ""
        try:
            # Common
            with open(common_style_path, "r") as common:
                stylesheet += common.read()

            # Highdpi
            if self.is_highdpi:
                with open(highdpi_style_path, "r") as highdpi:
                    stylesheet += highdpi.read()

            # Selected
            with open(selected_style_path, "r") as selected:
                stylesheet += selected.read()

        except IOError:
            pass

        cvlab_dir = os.path.abspath(__file__ + "/../../").replace("\\","/")
        stylesheet = stylesheet.replace("$CVLAB_DIR", cvlab_dir)

        # Switch background for highdpi
        if self.is_highdpi:
            stylesheet = stylesheet.replace("background.png", "background-highdpi.png")

        self.stylesheet = stylesheet
        self.main_window.setStyleSheet(stylesheet)
        refresh_style_recursive(self.main_window)

        # Load wires style
        self.wire_style = WireStyle(stylesheet)

        # Notify
        self.style_changed.emit()

        # Update settings
        current_style = self.main_window.settings.get_with_default(config.VIEW_SECTION, config.STYLE)
        if style_name == config.DEFAULTS[config.VIEW_SECTION][config.STYLE]:
            if self.main_window.settings.has_option(config.VIEW_SECTION, config.STYLE):
                self.main_window.settings.remove_option(config.VIEW_SECTION, config.STYLE)
        elif style_name != current_style:
            self.main_window.settings.set(config.VIEW_SECTION, config.STYLE, style_name)

    def apply_default_stylesheet(self):
        current_style = self.main_window.settings.get_with_default(config.VIEW_SECTION, config.STYLE)
        if current_style is not None:
            self.apply_stylesheet(current_style)

    def get_available_stylesheets(self):
        # Todo: fill this basing on 'themes' directory content
        return ["default", "dark"]


def refresh_style_recursive(base_widget):
    widgets = [base_widget]
    while widgets:
        widget = widgets.pop()
        base_widget.style().polish(widget)
        for child in widget.children():
            if isinstance(child, QWidget):
                widgets.append(child)



from .base import *


class Forwarder(NormalElement):
    name = "Forwarder"
    comment = "Forwards its input without processing"

    def __init__(self):
        super(Forwarder, self).__init__()

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               []

    def get_processing_units(self, inputs, parameters):
        outputs = {"output": inputs["input"]}
        units = [ProcessingUnit(self, inputs, parameters, outputs)]
        return units, outputs

    def process_units(self):
        pass

    def get_source(self):
        name = self.__class__.__name__.lower()
        src = """\
def {name}(inputs, outputs, parameters):
    outputs['output'] = inputs['input']
""".format(**locals())
        return name, src, []



class Sequencer(NormalElement):
    name = "Sequencer"
    comment = "Makes a single sequenced data from its inputs"

    def __init__(self):
        super(Sequencer, self).__init__()

    def get_attributes(self):
        return [Input("inputs", multiple=True)], \
               [Output("output", "Sequence")], \
            []

    def get_processing_units(self, inputs, parameters):
        outputs = {"output": Sequence(inputs["inputs"].value)}
        units = [ProcessingUnit(self, inputs, parameters, outputs)]
        return units, outputs

    def process_units(self):
        pass


class Desequencer(NormalElement):
    name = "Desequencer"
    comment = "Gets single sequenced data and separates it"

    def get_attributes(self):
        return [Input("input")], [Output("outputs", desequencing=True)], []

    def get_processing_units(self, inputs, parameters):
        outputs = {"outputs": inputs["input"]}
        units = [ProcessingUnit(self, inputs, parameters, outputs)]
        return units, outputs

    def process_units(self):
        pass





class Zip(NormalElement):
    name = "Zip"
    comment = "Zips its inputs"

    def get_attributes(self):
        return [Input("inputs", multiple=True)], [Output("outputs")], []

    def get_processing_units(self, inputs, parameters):
        input_list = inputs['inputs'].value

        if len(input_list) == 1 and input_list[0].type() == Data.SEQUENCE:
            input_list = input_list[0].value

        output_list = list(zip(*input_list))
        output_list = [Sequence(list(values)) for values in output_list]

        outputs = {"outputs": Sequence(output_list)}
        units = [ProcessingUnit(self, inputs, parameters, outputs)]

        return units, outputs

    def process_units(self):
        pass



class SequenceSelector(NormalElement):
    name = "Sequence selector"
    comment = "Selects a simple image from a sequence"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [IntParameter("number", value=0, min_=0, max_=100)]

    def get_processing_units(self, inputs, parameters):
        sequence = inputs["input"]
        if sequence.type() != Data.SEQUENCE or not sequence.value:
            raise TypeError("Input data must be a non-empty sequence")
        data = sequence.value[0]
        if not all(d.is_compatible(data) for d in sequence.value):
            raise TypeError("All data values in sequence must be compatible")
        outputs = {"output": data.create_placeholder()}
        units = [ProcessingUnit(self, inputs, parameters, outputs)]
        return units, outputs

    def process_inputs(self, inputs, outputs, parameters):
        outputs["output"] = inputs["input"].value[parameters["number"]]


class SequenceDeleter(NormalElement):
    name = "Sequence deleter"
    comment = "Deletes a simple image from a sequence"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [IntParameter("number", value=0, min_=0, max_=100)]

    def get_processing_units(self, inputs, parameters):
        sequence = inputs["input"]
        if sequence.type() != Data.SEQUENCE or not sequence.value or len(sequence.value) <= 1:
            raise TypeError("Input data must be a non-empty sequence")
        if not all(d.is_compatible(sequence.value[0]) for d in sequence.value):
            raise TypeError("All data values in sequence must be compatible")
        data = Sequence(sequence.value[:-1])
        outputs = {"output": data.create_placeholder()}
        units = [ProcessingUnit(self, inputs, parameters, outputs)]
        return units, outputs

    def process_inputs(self, inputs, outputs, parameters):
        deleted = parameters["number"]
        new_seq = inputs["input"].value[:]
        del new_seq[deleted]
        outputs["output"] = Sequence(new_seq)



class ConcatenateOperator(MultiInputOneOutputElement):
    name = "Concatenate operator"
    comment = "Concatenates input arrays"
    package = "Matrix miscellaneous"

    def process_inputs(self, inputs, outputs, parameters):
        output = Data()
        output.value = np.concatenate([d.value for d in inputs.values()])
        outputs["output"] = output


register_elements_auto(__name__, locals(), "Data flow", 1)


from .base import *


class OpenCVCanny(NormalElement):
    name = "Canny transform"
    comment = "Canny edge detector"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [IntParameter("thr1", "Threshold 1", min_=0, max_=255),
                IntParameter("thr2", "Threshold 2", min_=0, max_=255)]

    def process_inputs(self, inputs, outputs, parameters):
        image = inputs["input"].value

        color = False
        if len(image.shape) >= 3:
            image = cv.cvtColor(image, cv.COLOR_BGR2GRAY)
            color = True

        output = cv.Canny(image, parameters["thr1"], parameters["thr2"])

        if color:
            output = cv.cvtColor(output, cv.COLOR_GRAY2BGR)

        outputs["output"] = Data(output)


register_elements_auto(__name__, locals(), "Features 2D", 5)



from .base import *


class Rotate(NormalElement):
    name = "Rotate"
    comment = "Rotates using multiple 90"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [
                    IntParameter("angle", "Angle", 0, min_=0, max_=270, step=90),
               ]

    def process_inputs(self, inputs, outputs, parameters):
        image = np.copy(inputs["input"].value)
        angle = parameters["angle"]
        angle_rad = np.radians(angle)
        h, w = image.shape[:2]

        trans1 = np.array([[1, 0, -w / 2],
                           [0, 1, -h / 2],
                           [0, 0, 1]],
                          dtype=np.float32)
        if angle % 180 == 90:
            h, w = w, h
        trans2 = np.array([[1, 0, w / 2],
                           [0, 1, h / 2],
                           [0, 0, 1]],
                          dtype=np.float32)
        rot = np.array([[np.cos(angle_rad), -np.sin(angle_rad), 0],
                        [np.sin(angle_rad), np.cos(angle_rad), 0],
                        [0, 0, 1]],
                          dtype=np.float32)

        mat = np.dot(np.dot(trans2, rot), trans1)

        out = cv.warpAffine(image, mat[:2], (w, h))

        outputs["output"] = Data(out)


register_elements_auto(__name__, locals(), "Transforms", 5)


import random
import string
from threading import RLock
from weakref import WeakValueDictionary


objects = WeakValueDictionary()
lock = RLock()


def next_id(object2add=None):
    with lock:
        if not objects:
            id = 1
        else:
            id = max(objects.keys()) + 1
        if object2add is not None: objects[id] = object2add
        return id


def change_id(old, new):
    with lock:
        if old not in objects: return
        if new in objects: raise ValueError("ID actualy exists in ID manager")
        objects[new] = objects[old]
        objects.pop(old)


# sure, it's pseudo unique id
def unique_id(digits=6):
    valid_chars = list(set(string.digits.lower()))
    entropy_string = ''
    for i in range(digits):
        entropy_string += random.choice(valid_chars)
    return entropy_string



from ..diagram.diagram import Diagram
from ..diagram.interface import *


class Notifier:
    def __init__(self):
        self._lock = threading.Condition()
        self._do_notify = False

    def notify(self):
        with self._lock:
            self._do_notify = True
            self._lock.notifyAll()

    def wait(self):
        with self._lock:
            while not self._do_notify:
                self._lock.wait()
            self._do_notify = False

    def is_set(self):
        return self._do_notify


class Hook(QObject, object):
    """Hook for joining diagram elements together"""

    def __init__(self, connector):
        super(Hook, self).__init__()
        self.connector = connector
        self.lock = RLock()


class InputHook(Hook):
    """Hook for getting some input"""

    def __init__(self, connector):
        super(InputHook, self).__init__(connector)
        assert isinstance(self.connector, Input)
        self.sequence_indices = OrderedDict()
        if not self.connector.optional:
            self.empty_data = Sequence() if self.connector.multiple else EmptyData()
        else:
            self.empty_data = EmptyOptionalData()
        self.data = self.empty_data

    def get_data(self):
        return self.data

    def set_data(self, data, from_hook):
        assert data is not None
        assert from_hook is not None
        if self.connector.multiple:
            with self.lock:
                assert from_hook in self.sequence_indices
                index = self.sequence_indices[from_hook]
                if data is self.data.value[index]:
                    return
                self.data.value[index] = data
        else:
            with self.lock:
                if self.data is data: return
                self.data = data
        self.connector.parent.recalculate(False, True, True)

    def delete(self):
        pass

    def connected(self, from_hook):
        assert from_hook is not None
        if self.connector.multiple:
            with self.lock:
                assert from_hook not in self.sequence_indices
                self.sequence_indices[from_hook] = len(self.data.value)
                self.data = Sequence(self.data.value + [EmptyData()])
        # self.connector.parent.recalculate(False, True, True)

    def disconnected(self, from_hook):
        if self.connector.multiple:
            with self.lock:
                assert from_hook in self.sequence_indices
                del self.sequence_indices[from_hook]
                self.data = Sequence([self.data.value[i] for i in self.sequence_indices.values()])
                for i, hook in enumerate(self.sequence_indices.keys()):
                    self.sequence_indices[hook] = i
        else:
            self.data = self.empty_data
        self.connector.parent.recalculate(False, True, True)


class OutputHook(Hook):
    """Hook for giving output"""

    def __init__(self, connector):
        super(OutputHook, self).__init__(connector)
        assert isinstance(self.connector, Output)
        self.empty_data = Data(None, Data.NONE)
        self.data = self.empty_data

    # gui thread only
    # @pyqtSlot(InputHook)
    def connected(self, to_hook):
        self.actualize_outputs()  # fixme: to jest chyba nadmiarowe...

    # gui thread only
    # @pyqtSlot(InputHook)
    def disconnected(self, to_hook):
        self.actualize_outputs()  # fixme: to jest chyba nadmiarowe...

    # element thread only
    # @pyqtSlot(object)
    def set_data(self, data):
        with self.lock:
            self.data = data if data is not None else self.empty_data
        self.actualize_outputs()

    # gui thread or element thread
    # @pyqtSlot()
    def actualize_outputs(self):
        with Diagram.diagram_lock.reader, self.lock:
            if self.connector.desequencing and self.data and self.data.type() == Data.SEQUENCE and len(
                    self.connector.connected_to) > 1:
                for input, data in zip(self.connector.connected_to, self.data.value):
                    input.hook.set_data(data, self)
            else:
                for i in self.connector.connected_to:
                    i.hook.set_data(self.data, self)

    # gui thread or other element thread
    def get_data(self):
        return self.data


class InputQtHook(InputHook):
    set_data_signal = pyqtSignal(object, OutputHook)

    def __init__(self, connector):
        super(InputQtHook, self).__init__(connector)
        self.set_data_signal.connect(super(InputQtHook, self).set_data)

    def set_data(self, data, from_hook):
        # print "set_data signal emit from", thread.get_ident()
        self.set_data_signal.emit(data, from_hook)



class OutputQtHook(OutputHook):
    set_data_signal = pyqtSignal(object)
    outdate_signal = pyqtSignal()
    actualize_outputs_signal = pyqtSignal()

    def __init__(self, connector):
        super(OutputQtHook, self).__init__(connector)
        self.set_data_signal.connect(super(OutputQtHook, self).set_data)
        self.outdate_signal.connect(super(OutputQtHook, self).outdate)
        self.actualize_outputs_signal.connect(super(OutputQtHook, self).actualize_outputs)

    def set_data(self, data):
        self.set_data_signal.emit(data)

    def outdate(self):
        self.outdate_signal.emit()

    def actualize_outputs(self):
        self.actualize_outputs_signal.emit()




import os
if os.name != 'nt': raise ImportError("Kinect module only work on Windows")

from datetime import datetime, timedelta, time
from threading import Event

from cvlab.diagram.elements.base import *

from .thirdparty.pykinect import nui


class Kinect(NormalElement):
    name = "Kinect (pykinect)"
    comment = "Reads data from Kinect device"
    device_lock = Lock()

    def __init__(self):
        super(Kinect, self).__init__()
        self.last_frame_time = datetime.now()
        self.play = Event()  # todo: ladowanie tego powinno byc z parametru - zapamietywane przy zapisie/odczycie
        self.play.set()
        self.last_parameters = None
        self.kinect = None
        self.color_frame = None
        self.depth_frame = None
        self.color_timestamp = None
        self.depth_timestamp = None
        self.color_image = None
        self.depth_image = None
        self.kinect_color_lock = threading.Lock()
        self.kinect_depth_lock = threading.Lock()
        self.recalculate(True, True, True)

    def get_attributes(self):
        return [], \
               [Output("color"), Output("depth")], \
               [IntParameter("device", value=0, min_=0, max_=10),
                ComboboxParameter("color res", [("640x480", 0), ("1280x960", 1)]),
                ComboboxParameter("depth res", [("640x480", 0), ("320x240", 1)]),
                FloatParameter("fps", value=15, min_=0.1, max_=120),
                ButtonParameter("pause", self.playpause, "Play / Pause")]

    def playpause(self):
        if self.play.is_set():
            self.play.clear()
        else:
            self.play.set()

    def delete(self):
        self.play.set()
        ThreadedElement.delete(self)
        if self.kinect is not None:
            self.kinect.close()

    def process(self):
        print("process")
        if not self.check_parameters_up_to_date():
            if self.kinect is not None:
                self.kinect.close()
                self.kinect = None
            try:
                self.kinect = nui.Runtime(index=self.parameters["device"].get())
            except:
                raise Exception("Device not connected or Microsoft Kinect SDK missing")
            self.kinect.video_frame_ready += self.video_frame_ready
            self.kinect.depth_frame_ready += self.depth_frame_ready

            color_resolution = self.parameters["color res"].get()
            if color_resolution == 0:
                res = nui.ImageResolution.Resolution640x480
            else:
                res = nui.ImageResolution.Resolution1280x1024
            self.kinect.video_stream.open(nui.ImageStreamType.Video, 2, res, nui.ImageType.Color)

            depth_resolution = self.parameters["depth res"].get()
            if depth_resolution == 0:
                res = nui.ImageResolution.Resolution640x480
            else:
                res = nui.ImageResolution.Resolution320x240
            self.kinect.depth_stream.open(nui.ImageStreamType.Depth, 2, res, nui.ImageType.Depth)

        if self.kinect is not None:
            while True:
                self.produce_output()

    def check_parameters_up_to_date(self):
        parameters = self.get_the_parameters()
        result = self.last_parameters == parameters
        self.last_parameters = parameters
        return result

    def get_the_parameters(self):
        parameters = {}
        for name in ["device", "color res", "depth res"]:
            parameters[name] = self.parameters[name].get()
        return parameters

    def video_frame_ready(self, frame):
        with self.kinect_color_lock:
            self.color_frame = frame
            self.color_image = self.get_color_image()
            self.color_timestamp = frame.timestamp

    def depth_frame_ready(self, frame):
        with self.kinect_depth_lock:
            self.depth_frame = frame
            self.depth_image = self.get_depth_image()
            self.depth_timestamp = frame.timestamp

    def produce_output(self):
        print("produce")
        self.may_interrupt()
        self.sleep_if_needed()
        self.may_interrupt()

        update = False
        color = None
        depth = None
        with self.kinect_color_lock:
            with self.kinect_depth_lock:
                if self.color_image is not None and self.depth_image is not None and self.timestamps_ok():
                    # copy with eliminating mirroring
                    color = self.color_image[:, ::-1, :]
                    depth = self.depth_image[:, ::-1]
                    # get rid of player indexes
                    depth = np.right_shift(depth, 3)
                    update = True

        self.may_interrupt()
        if update:
            self.may_interrupt()
            self.set_state(Element.STATE_BUSY)
            if color is not None and len(color) > 0 and depth is not None and len(depth) > 0:
                self.outputs["color"].put(Data(color))
                self.outputs["depth"].put(Data(depth))
                self.last_frame_time = datetime.now()
                self.set_state(Element.STATE_READY)
                self.notify_state_changed()
        self.may_interrupt()
        self.play.wait()

    def timestamps_ok(self):
        return abs(self.color_timestamp - self.depth_timestamp) <= 10

    def sleep_if_needed(self):
        now = datetime.now()
        if now - self.last_frame_time < timedelta(seconds=1.0 / self.parameters["fps"].value):
            seconds_to_wait = 1.0 / self.parameters["fps"].value - (now - self.last_frame_time).total_seconds()
            breaks = int(round(seconds_to_wait * 10 + 1))
            for _ in range(breaks):
                time.sleep(seconds_to_wait / breaks)
                self.may_interrupt()

    def get_color_image(self):
        if self.color_frame is not None:
            image = self.prepare_color_image()
            self.color_frame.image.copy_bits(image.ctypes.data)
            return image[:, :, :3]

    def get_depth_image(self):
        if self.depth_frame is not None:
            depth = self.prepare_depth_image()
            self.depth_frame.image.copy_bits(depth.ctypes.data)
            return depth

    def prepare_color_image(self):
        if self.color_frame.resolution == nui.ImageResolution.Resolution1280x1024:
            return np.zeros((1024, 960, 4), np.uint8)
        else:
            return np.zeros((480, 640, 4), np.uint8)

    def prepare_depth_image(self):
        if self.depth_frame.resolution == nui.ImageResolution.Resolution640x480:
            return np.zeros((480, 640, 1), np.uint16)
        else:
            return np.zeros((240, 320, 1), np.uint16)


register_elements_auto(__name__, locals(), "Kinect", 10)




class Input:
    def __init__(self, id, name=None, multiple=False, optional=False):
        super(Input, self).__init__()
        self.id = id
        self.parent = None
        if name is None: name = id
        self.name = name
        self.multiple = multiple
        self.optional = optional
        self.connected_from = []
        self.hook = None
        from .diagram import Diagram
        self.diagram_write_lock = Diagram.diagram_lock.writer

    def connect(self, output):
        with self.diagram_write_lock:
            if output in self.connected_from: return
            if not self.multiple:
                self.disconnect_all()
            self.connected_from.append(output)
            self.hook.connected(output.hook)

    def disconnect(self, output):
        with self.diagram_write_lock:
            if output not in self.connected_from: return
            self.connected_from.remove(output)
            output.disconnect(self)
            self.hook.disconnected(output.hook)
            self.parent.diagram.notify_disconnect(output, self)

    def disconnect_all(self):
        with self.diagram_write_lock:
            for output in list(self.connected_from):
                self.disconnect(output)

    def get(self):
        return self.hook.get_data()


class Output:
    def __init__(self, id, name=None, desequencing=False):
        super(Output, self).__init__()
        self.id = id
        self.parent = None
        if name is None: name = id
        self.name = name
        self.desequencing = desequencing
        self.connected_to = []
        self.hook = None
        from .diagram import Diagram
        self.diagram_write_lock = Diagram.diagram_lock.writer

    def connect(self, input_):
        with self.diagram_write_lock:
            if input_ in self.connected_to: return
            self.connected_to.append(input_)
            self.hook.connected(input_.hook)

    def disconnect(self, input_):
        with self.diagram_write_lock:
            if input_ not in self.connected_to: return
            self.connected_to.remove(input_)
            input_.disconnect(self)
            self.hook.disconnected(input_.hook)

    def disconnect_all(self):
        with self.diagram_write_lock:
            for input_ in list(self.connected_to):
                self.disconnect(input_)

    def put(self, data):
        self.hook.set_data(data)

    def get(self):
        return self.hook.get_data()





from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import *


class TabsContainer(QTabWidget):

    def __init__(self):
        super(TabsContainer, self).__init__()
        self.diagram_manager = None
        self.tab_manager = TabManager(self.tabBar())
        self.setTabShape(QTabWidget.Triangular)
        self.tabBar().setContextMenuPolicy(Qt.CustomContextMenu)
        self.tabBar().customContextMenuRequested.connect(self.tab_manager.show_context_menu)


class TabManager:
    def __init__(self, tab_bar):
        self.tab_bar = tab_bar

    def show_context_menu(self, point):
        if not point:
            return
        idx = self.tab_bar.tabAt(point)
        if idx == -1:
            return
        menu = QMenu(self.tab_bar)
        menu.addAction(get_action(self.tab_bar, TabAction.CLOSE, idx))
        menu.addSeparator()
        menu.addAction(get_action(self.tab_bar, TabAction.SAVE, idx))
        menu.addAction(get_action(self.tab_bar, TabAction.SAVE_AS, idx))
        menu.exec_(self.tab_bar.mapToGlobal(point))


def get_action(parent, action_type, idx):
        if action_type == TabAction.CLOSE:
            action = TabAction("Close", parent, idx)
            action.triggered.connect(action.close_diagram)
            return action
        elif action_type == TabAction.SAVE:
            action = TabAction("Save", parent, idx)
            action.triggered.connect(action.save_diagram)
            return action
        elif action_type == TabAction.SAVE_AS:
            action = TabAction("Save as...", parent, idx)
            action.triggered.connect(action.save_diagram_as)
            return action


class TabAction(QAction):

    CLOSE = 0
    SAVE = 1
    SAVE_AS = 2

    def __init__(self, text, parent, tab_idx):
        super(TabAction, self).__init__(text, parent)
        self.tab_idx = tab_idx

    def get_diagram_manager(self):
        return self.parent().parent().diagram_manager

    def close_diagram(self):
        self.get_diagram_manager().close_diagram(self.tab_idx)

    def save_diagram(self):
        self.get_diagram_manager().save_diagram(self.tab_idx)

    def save_diagram_as(self):
        self.get_diagram_manager().save_diagram_as(self.tab_idx)



 # ############################################################################
 #
 # Copyright (c) Microsoft Corporation. 
 #
 # Available under the Microsoft PyKinect 1.0 Alpha license.  See LICENSE.txt
 # for more information.
 #
 # ###########################################################################/


import ctypes
import os
import _thread

# basic initialization, Python specific infrastructure
_nuidll_path = os.path.join(os.environ['WINDIR'], 'System32', 'Kinect10.dll')
_NUIDLL = ctypes.WinDLL(_nuidll_path)

class KinectError(WindowsError):
    """Represents an error from a Kinect sensor"""
    pass

from thirdparty.pykinect.nui.structs import (ImageDigitalZoom, ImageFrame, ImageResolution,
                                  ImageType, ImageViewArea, JointId, 
                                  JointTrackingState, PlanarImage, SkeletonData, 
                                  SkeletonFrame, SkeletonFrameQuality, 
                                  SkeletonQuality, SkeletonTrackingState, 
                                  TransformSmoothParameters, Vector, _Enumeration)

from ._interop import (_CreateEvent, _CloseHandle, _WaitForSingleObject, 
                      _WaitForMultipleObjects, _WAIT_OBJECT_0, _INFINITE, 
                      _SysFreeString, _NuiInstance, _NuiCreateSensorByIndex, 
                      _NuiGetSensorCount)


_NUI_IMAGE_PLAYER_INDEX_SHIFT      =    3
_NUI_IMAGE_PLAYER_INDEX_MASK       =    ((1 << _NUI_IMAGE_PLAYER_INDEX_SHIFT)-1)
_NUI_IMAGE_DEPTH_MAXIMUM           =    ((4000 << _NUI_IMAGE_PLAYER_INDEX_SHIFT) | _NUI_IMAGE_PLAYER_INDEX_MASK)
_NUI_IMAGE_DEPTH_MINIMUM           =    (800 << _NUI_IMAGE_PLAYER_INDEX_SHIFT)
_NUI_IMAGE_DEPTH_NO_VALUE          =    0

_NUI_CAMERA_DEPTH_NOMINAL_FOCAL_LENGTH_IN_PIXELS         = 285.63   # Based on 320x240 pixel size.
_NUI_CAMERA_DEPTH_NOMINAL_INVERSE_FOCAL_LENGTH_IN_PIXELS = 3.501e-3 # (1/NUI_CAMERA_DEPTH_NOMINAL_FOCAL_LENGTH_IN_PIXELS)
_NUI_CAMERA_DEPTH_NOMINAL_DIAGONAL_FOV                   = 70.0
_NUI_CAMERA_DEPTH_NOMINAL_HORIZONTAL_FOV                 = 58.5
_NUI_CAMERA_DEPTH_NOMINAL_VERTICAL_FOV                   = 45.6

_NUI_CAMERA_COLOR_NOMINAL_FOCAL_LENGTH_IN_PIXELS         = 531.15   # Based on 640x480 pixel size.
_NUI_CAMERA_COLOR_NOMINAL_INVERSE_FOCAL_LENGTH_IN_PIXELS = 1.83e-3  # (1/NUI_CAMERA_COLOR_NOMINAL_FOCAL_LENGTH_IN_PIXELS)
_NUI_CAMERA_COLOR_NOMINAL_DIAGONAL_FOV                   = 73.9
_NUI_CAMERA_COLOR_NOMINAL_HORIZONTAL_FOV                 = 62.0
_NUI_CAMERA_COLOR_NOMINAL_VERTICAL_FOV                   = 48.6

# the max # of NUI output frames you can hold w/o releasing
_NUI_IMAGE_STREAM_FRAME_LIMIT_MAXIMUM = 4

# return S_FALSE instead of E_NUI_FRAME_NO_DATA if NuiImageStreamGetNextFrame( ) doesn't have a frame ready and a timeout != INFINITE is used
_NUI_IMAGE_STREAM_FLAG_SUPPRESS_NO_FRAME_DATA  = 0x00010000

#######################################################

_NUI_SKELETON_MAX_TRACKED_COUNT = 2
_NUI_SKELETON_INVALID_TRACKING_ID = 0

# Assuming a pixel resolution of 320x240
# x_meters = (x_pixelcoord - 160) * NUI_CAMERA_DEPTH_IMAGE_TO_SKELETON_MULTIPLIER_320x240 * z_meters;
# y_meters = (y_pixelcoord - 120) * NUI_CAMERA_DEPTH_IMAGE_TO_SKELETON_MULTIPLIER_320x240 * z_meters;
_NUI_CAMERA_DEPTH_IMAGE_TO_SKELETON_MULTIPLIER_320x240 = _NUI_CAMERA_DEPTH_NOMINAL_INVERSE_FOCAL_LENGTH_IN_PIXELS
 
# Assuming a pixel resolution of 320x240
# x_pixelcoord = (x_meters) * NUI_CAMERA_SKELETON_TO_DEPTH_IMAGE_MULTIPLIER_320x240 / z_meters + 160;
# y_pixelcoord = (y_meters) * NUI_CAMERA_SKELETON_TO_DEPTH_IMAGE_MULTIPLIER_320x240 / z_meters + 120;
_NUI_CAMERA_SKELETON_TO_DEPTH_IMAGE_MULTIPLIER_320x240 = _NUI_CAMERA_DEPTH_NOMINAL_FOCAL_LENGTH_IN_PIXELS

_FLT_EPSILON     = 1.192092896e-07

class RuntimeOptions(object):
    """Specifies the runtime options for a Kinect sensor. """
    uses_depth_and_player_index = UseDepthAndPlayerIndex = 0x01
    uses_color = UseColor = 0x02
    uses_skeletal_tracking = UseSkeletalTracking = 0x08
    uses_depth = UseDepth = 0x20
    uses_high_quality_color = 0x40   # implies COLOR stream will be from uncompressed YUY2 @ 15fps
    uses_audio = UsesAudio = 0x10000000


class Device(object):
    """Represents a system's Kinect sensors."""
    _device_inst = None

    def __new__(cls):
        if Device._device_inst is None:
            Device._device_inst = object.__new__(Device)
        return Device._device_inst
        
    @property
    def count(self):
        """The number of active Kinect sensors that are attached to the system."""
        return _NuiGetSensorCount()


class Runtime(object):
    """Represents a Kinect sensor."""

    def __init__(self, 
                 nui_init_flags = RuntimeOptions.uses_color| RuntimeOptions.uses_depth | RuntimeOptions.uses_depth_and_player_index | RuntimeOptions.uses_skeletal_tracking,                 
                 index = 0):
        """Creates a new runtime.  By default initialized to the 1st installed kinect device and tracking all events"""
        self._nui = self._skeleton_event = self._image_event = self._depth_event = None
        self._nui = _NuiCreateSensorByIndex(index)
        try:            
            self._nui.NuiInitialize(nui_init_flags)
        except:
            self._nui.NuiShutdown()
            import traceback
            
            raise KinectError('Unable to create Kinect runtime '+ traceback.format_exc()) 

        self.depth_frame_ready = _event()
        self.skeleton_frame_ready = _event()
        self.video_frame_ready = _event()

        self._skeleton_event = _CreateEvent(None, True, False, None)
        self._image_event = _CreateEvent(None, True, False, None)
        self._depth_event = _CreateEvent(None, True, False, None)
        
        self.camera = Camera(self)
        self.skeleton_engine = SkeletonEngine(self)
        self.depth_stream = ImageStream(self)
        self.video_stream = ImageStream(self)
        
        _thread.start_new_thread(self._event_thread, ())

    def close(self):
        """closes the current runtime"""
        if self._skeleton_event is not None:
            _CloseHandle(self._skeleton_event)
            self._skeleton_event = None
        
        if self._image_event is not None:
            _CloseHandle(self._image_event)
            self._image_event = None
        
        if self._depth_event is not None:
            _CloseHandle(self._depth_event)
            self._depth_event = None

        if self._nui is not None:
            self._nui.NuiShutdown()
            self._nui = None

    def _check_closed(self):
        if self._nui is None:
            raise KinectError('Device closed')

    def __del__(self):
        self.close()
    
    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    @property
    def instance_index(self):
        self._check_closed()

        """Gets the index for this instance of Runtime."""
        return self._nui.InstanceIndex()


    def _event_thread(self):
        handles = (ctypes.c_voidp * 3)()
        handles[0] = self._skeleton_event
        handles[1] = self._depth_event
        handles[2] = self._image_event
        while 1:    
            wait = _WaitForMultipleObjects(3, handles, False, _INFINITE)
            if wait == 0:
                # skeleton data                
                try:
                    frame = self._nui.NuiSkeletonGetNextFrame(0)
                except KinectError:
                    continue
        
                for curSkeleton in frame.SkeletonData:
                    if curSkeleton.eTrackingState != SkeletonTrackingState.NOT_TRACKED:
                        self.skeleton_frame_ready.fire(frame)
                        break
            elif wait == 1:
                # depth event
                depth_frame = self._nui.NuiImageStreamGetNextFrame(self.depth_stream._stream, 0)
                self.depth_frame_ready.fire(depth_frame)                
                self._nui.NuiImageStreamReleaseFrame(self.depth_stream._stream, depth_frame)
            elif wait == 2:
                # image event
                depth_frame = self._nui.NuiImageStreamGetNextFrame(self.video_stream._stream, 0)
                self.video_frame_ready.fire(depth_frame)     
                self._nui.NuiImageStreamReleaseFrame(self.video_stream._stream, depth_frame)
                pass
            else:
                # wait failed in some form (abandoned, timeout, or failed), this ends our loop
                # when we close our events.
                break


class ImageStreamType(object):
    """Specifies an image stream type. """
    depth = Depth = 0
    video = Video = 1
    invalid = Invalid = -1


class ImageStream(object):
    """Represents an image stream."""

    def __init__(self, runtime):
        self.runtime = runtime
        self.resolution = ImageResolution.Invalid
        self.height = self.width = 0
        self.stream_type = ImageStreamType.Invalid
        self._stream = None

    def open(self, image_stream_type = 0, frame_limit = 2, resolution = ImageResolution.Resolution320x240, image_type = ImageType.Color):
        if image_stream_type == ImageStreamType.Depth:
            event_handle = self.runtime._depth_event
        elif image_stream_type == ImageStreamType.Video:
            event_handle = self.runtime._image_event
        else:
            raise ValueError("Unexpected image stream type: %r" % (image_stream_type, ))

        if resolution == ImageResolution.Resolution1280x1024:
            self.width = 1280
            self.height = 1024
        elif resolution == ImageResolution.Resolution640x480:
            self.width = 640
            self.height = 480
        elif resolution == ImageResolution.Resolution320x240:
            self.width = 320
            self.height = 240
        elif resolution == ImageResolution.Resolution80x60:
            self.width = 80
            self.height = 60
        else:
            raise ValueError("Unexpected resolution: %r" % (resolution, ))

        self._stream = self.runtime._nui.NuiImageStreamOpen(image_type, resolution, 0, frame_limit, event_handle)
        self.stream_type = image_stream_type
        self.resolution = resolution        
        self.type = image_type

    def get_next_frame(self, milliseconds_to_wait = 0):
        # TODO: Allow user to provide a NUI_IMAGE_FRAME ?
        return self.runtime._nui.NuiImageStreamGetNextFrame(self._stream, milliseconds_to_wait)
    
    @staticmethod
    def get_valid_resolutions(image_type):
        if image_type == ImageType.Color:
            return (ImageResolution.Resolution1280x1024, ImageResolution.Resolution640x480)
        elif image_type == ImageType.Depth:
            return (ImageResolution.Resolution640x480, )
        elif image_type == ImageType.DepthAndPlayerIndex:
            return (ImageResolution.Resolution320x240, )
        elif image_type == ImageType.ColorYuv:
            return (ImageResolution.Resolution640x480, )
        elif image_type == ImageType.ColorYuvRaw:
            return (ImageResolution.Resolution640x480, )
        else:
            raise KinectError("Unknown image_type: %r" % (image_type, ))


class SkeletonEngine(object):    
    """Represents the skeleton tracking engine. """

    def __init__(self, runtime):
        self.runtime = runtime
        self._enabled = False

    def get_enabled(self):
        return self._enabled

    def set_enabled(self, value):
        if value:
            self.runtime._nui.NuiSkeletonTrackingEnable(self.runtime._skeleton_event)
            self._enabled = True
        else:
            self.runtime._nui.NuiSkeletonTrackingDisable(self.runtime._skeleton_event)
            self._enabled = False

    enabled = property(get_enabled, set_enabled)

    def get_next_frame(self, timeout = -1):
        res = self.runtime._nui.NuiSkeletonGetNextFrame(timeout)
        assert isinstance(res, SkeletonFrame)
        return res

    @staticmethod
    def depth_image_to_skeleton(fDepthX, fDepthY, usDepthValue):
        """returns Vector4"""

        ##
        ##  Depth is in meters in skeleton space.
        ##  The depth image pixel format has depth in millimeters shifted left by 3.
        ##
    
        fSkeletonZ = (usDepthValue >> 3)/1000.0
    
        ##
        ## Center of depth sensor is at (0,0,0) in skeleton space, and
        ## and (160,120) in depth image coordinates.  Note that positive Y
        ## is up in skeleton space and down in image coordinates.
        ##
    
        fSkeletonX = (fDepthX - 0.5) * (_NUI_CAMERA_DEPTH_IMAGE_TO_SKELETON_MULTIPLIER_320x240 * fSkeletonZ) * 320.0
        fSkeletonY = (0.5 - fDepthY) * (_NUI_CAMERA_DEPTH_IMAGE_TO_SKELETON_MULTIPLIER_320x240 * fSkeletonZ) * 240.0
    
        ##
        ## Return the result as a vector.
        ##
        
        v4 = Vector()    
        v4.x = fSkeletonX
        v4.y = fSkeletonY
        v4.z = fSkeletonZ
        v4.w = 1.0
        return v4

    @staticmethod
    def skeleton_to_depth_image(vPoint, scaleX = 1, scaleY = 1):
        """Given a Vector4 returns X and Y coordinates fo display on the screen.  Returns a tuple depthX, depthY"""

        if vPoint.z > _FLT_EPSILON: 
           ##
           ## Center of depth sensor is at (0,0,0) in skeleton space, and
           ## and (160,120) in depth image coordinates.  Note that positive Y
           ## is up in skeleton space and down in image coordinates.
           ##
       
           pfDepthX = 0.5 + vPoint.x * ( _NUI_CAMERA_SKELETON_TO_DEPTH_IMAGE_MULTIPLIER_320x240 / vPoint.z ) / 320.0
           pfDepthY = 0.5 - vPoint.y * ( _NUI_CAMERA_SKELETON_TO_DEPTH_IMAGE_MULTIPLIER_320x240 / vPoint.z ) / 240.0
       
           return pfDepthX * scaleX, pfDepthY * scaleY

        return 0.0, 0.0


class Camera(object):
    """Represents a Kinect sensor's camera."""

    def __init__(self, runtime):
        self.runtime = runtime
        self.elevation_angle

    ElevationMaximum = 27
    ElevationMinimum = -27

    def get_elevation_angle(self):
        """Gets or sets the camera elevation angle. """
        return self.runtime._nui.NuiCameraElevationGetAngle()

    def set_elevation_angle(self, degrees):
        """Gets or sets the camera elevation angle. """

        self.runtime._nui.NuiCameraElevationSetAngle(degrees)

    elevation_angle = property(get_elevation_angle, set_elevation_angle)

    @property
    def unique_device_name(self):
        """Gets the camera's unique device name. """
        return self.runtime._nui.GetUniqueDeviceName()

    def get_color_pixel_coordinates_from_depth_pixel(self, color_resolution, view_area, depth_x, depth_y, depth_value):
        """Returns the pixel coordinates in color image space that correspond to the specified pixel coordinates in depth image space. 

color_resolution: An ImageResolution value specifying the color image resolution.
view_area: An ImageViewArea structure containing the pan and zoom settings. If you provide this argument, you should pass in the view area from the image frame against which you are registering pixels, rather than manually instantiating and populating the structure. This helps ensure that your settings are valid.
depth_x: The x coordinate in depth image space.
depth_y: The y coordinate in depth image space.
depth_value The depth value in depth image space.

Returns: color_x, color_y - the x and y coordinate in the color image space
"""
        return self.runtime._nui.NuiImageGetColorPixelCoordinatesFromDepthPixel(
            color_resolution, 
            view_area, 
            depth_x, 
            depth_y, 
            depth_value)
        

def TransformSmoothParameters(vPoint):
    """returns depthX (float), depthY (float), depthValue (int)"""

    if vPoint.vector.z > _FLT_EPSILON:

       # Center of depth sensor is at (0,0,0) in skeleton space, and
       # and (160,120) in depth image coordinates.  Note that positive Y
       # is up in skeleton space and down in image coordinates.
       #
       
       pfDepthX = 0.5 + vPoint.vector.x * _NUI_CAMERA_SKELETON_TO_DEPTH_IMAGE_MULTIPLIER_320x240 / ( vPoint.vector.z * 320.0 )
       pfDepthY = 0.5 - vPoint.vector.y * _NUI_CAMERA_SKELETON_TO_DEPTH_IMAGE_MULTIPLIER_320x240 / ( vPoint.vector.z * 240.0 )
       
       #
       #  Depth is in meters in skeleton space.
       #  The depth image pixel format has depth in millimeters shifted left by 3.
       #
       
       pusDepthValue = int(vPoint.vector.z * 1000) << 3
       return pfDepthX, pfDepthY, pusDepthValue

    return 0.0, 0.0, 0


class _event(object):
    """class used for adding/removing/invoking a set of listener functions"""
    __slots__ = ['handlers']
        
    def __init__(self):
        self.handlers = []
    
    def __iadd__(self, other):
        self.handlers.append(other)
        return self
        
    def __isub__(self, other):
        self.handlers.remove(other)
        return self

    def fire(self, *args):
        for handler in self.handlers:
            handler(*args)


